# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from types import ModuleType

import numpy as np

from gt4py._core import definitions as core_defs
from gt4py.eve.extended_typing import NestedTuple
from gt4py.next import common, named_collections, utils
from gt4py.next.type_system import type_specifications as ts, type_translation


try:
    import cupy as cp
except ImportError:
    cp = None


@utils.tree_map
def asnumpy(field: common.Field | np.ndarray) -> np.ndarray:
    return field.asnumpy() if isinstance(field, common.Field) else field


def field_from_typespec(
    type_: ts.CollectionTypeSpec | ts.ScalarType, domain: common.Domain, xp: ModuleType
) -> common.MutableField | tuple[common.MutableField | tuple, ...]:
    """
    Allocate a field or (arbitrarily nested) tuple(s) of fields.

    The tuple structure and dtype is taken from a type_specifications.DataType,
    which is either ScalarType or a CollectionTypeSpec of ScalarType (possibly nested).

    >>> field_from_typespec(
    ...     ts.ScalarType(kind=ts.ScalarKind.INT32), common.domain({common.Dimension("I"): 1}), np
    ... )  # doctest: +ELLIPSIS
    NumPyArrayField(... dtype=int32...)
    >>> field_from_typespec(
    ...     ts.TupleType(
    ...         types=[
    ...             ts.ScalarType(kind=ts.ScalarKind.INT32),
    ...             ts.ScalarType(kind=ts.ScalarKind.FLOAT32),
    ...         ]
    ...     ),
    ...     common.domain({common.Dimension("I"): 1}),
    ...     np,
    ... )  # doctest: +ELLIPSIS
    (NumPyArrayField(... dtype=int32...), NumPyArrayField(... dtype=float32...))
    """

    def _constructor(
        type_: ts.CollectionTypeSpec,
        elems: NestedTuple[common.MutableField],
    ) -> named_collections.NamedCollection:
        if isinstance(type_, ts.NamedCollectionType):
            return named_collections.make_named_collection_constructor_from_type_spec(type_)(elems)
        return tuple(elems)

    @utils.tree_map(
        collection_type=ts.COLLECTION_TYPE_SPECS,
        result_collection_constructor=_constructor,
    )
    def impl(type_: ts.ScalarType) -> common.MutableField:
        res = common._field(
            xp.empty(domain.shape, dtype=xp.dtype(type_translation.as_dtype(type_).scalar_type)),
            domain=domain,
        )
        assert isinstance(res, common.MutableField)
        return res

    return impl(type_)


def get_array_ns(
    *args: core_defs.Scalar | common.Field | tuple[core_defs.Scalar | common.Field | tuple, ...],
) -> ModuleType:
    for arg in utils.flatten_nested_tuple(args):
        if hasattr(arg, "array_ns"):
            return arg.array_ns
    return np


def verify_device_field_type(field: common.Field, device: core_defs.DeviceType) -> bool:
    """Check if `field` is suitable for `device`."""
    if not (array_ns := getattr(field, "array_ns", False)):
        return False  # not a NDArrayField
    if device in [core_defs.DeviceType.CUDA, core_defs.DeviceType.ROCM]:
        assert core_defs.CUPY_DEVICE_TYPE is not None
        # TODO(havogt): generalize to other array libraries
        return device == core_defs.CUPY_DEVICE_TYPE and array_ns == cp
    else:
        return array_ns == np



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import copy
import os
import pathlib
import re
from typing import TYPE_CHECKING, ClassVar

from dace import SDFG, DebugInfo, Memlet, SDFGState, config, data, dtypes, nodes, subsets, symbolic
from dace.codegen import codeobject
from dace.sdfg.analysis.schedule_tree import treenodes as tn
from dace.sdfg.utils import inline_sdfgs

from gt4py._core import definitions as core_defs
from gt4py.cartesian import config as gt_config, definitions
from gt4py.cartesian.backend.base import register
from gt4py.cartesian.backend.gtc_common import (
    BackendCodegen,
    BaseGTBackend,
    CUDAPyExtModuleGenerator,
    GTBackendOptions,
    PyExtModuleGenerator,
    bindings_main_template,
    pybuffer_to_sid,
)
from gt4py.cartesian.backend.module_generator import make_args_data_from_gtir
from gt4py.cartesian.gtc import gtir
from gt4py.cartesian.gtc.dace import passes
from gt4py.cartesian.gtc.dace.oir_to_treeir import OIRToTreeIR
from gt4py.cartesian.gtc.dace.treeir_to_stree import TreeIRToScheduleTree
from gt4py.cartesian.gtc.dace.utils import array_dimensions, replace_strides
from gt4py.cartesian.gtc.definitions import CartesianSpace
from gt4py.cartesian.gtc.gtir_to_oir import GTIRToOIR
from gt4py.cartesian.gtc.passes.gtir_k_boundary import compute_k_boundary
from gt4py.cartesian.gtc.passes.oir_optimizations import caches
from gt4py.cartesian.gtc.passes.oir_optimizations.utils import compute_fields_extents
from gt4py.cartesian.gtc.passes.oir_pipeline import DefaultPipeline
from gt4py.cartesian.utils import shash
from gt4py.eve import codegen
from gt4py.eve.codegen import MakoTemplate as as_mako
from gt4py.storage.cartesian import layout


if TYPE_CHECKING:
    from gt4py.cartesian.stencil_builder import StencilBuilder
    from gt4py.cartesian.stencil_object import StencilObject


def _specialize_transient_strides(
    sdfg: SDFG,
    layout_info: layout.LayoutInfo,
    replacement_dictionary: dict[str, str] | None = None,
) -> None:
    # Find transients in this SDFG to specialize.
    stride_replacements = replace_strides(
        [
            array
            for array in sdfg.arrays.values()
            if isinstance(array, data.Array) and array.transient
        ],
        layout_info,
    )

    # In case of nested SDFGs (see below), merge with replacement dict that was passed down.
    # Dev note: We shouldn't use mutable data structures as argument defaults.
    replacement_dictionary = {} if replacement_dictionary is None else replacement_dictionary
    replacement_dictionary.update(stride_replacements)

    # Replace in this SDFG
    sdfg.replace_dict(replacement_dictionary)
    for state in sdfg.nodes():
        for node in state.nodes():
            if isinstance(node, nodes.NestedSDFG):
                # Recursively replace strides in nested SDFGs
                _specialize_transient_strides(node.sdfg, layout_info, replacement_dictionary)
    for k in replacement_dictionary.keys():
        if k in sdfg.symbols:
            sdfg.remove_symbol(k)


def _sdfg_add_arrays_and_edges(
    field_info: dict[str, definitions.FieldInfo],
    wrapper_sdfg: SDFG,
    state: SDFGState,
    inner_sdfg: SDFG,
    nsdfg: nodes.NestedSDFG,
    inputs: set[str] | dict[str, dtypes.typeclass],
    outputs: set[str] | dict[str, dtypes.typeclass],
    origins: dict[str, tuple[int, ...]],
) -> None:
    for name, array in inner_sdfg.arrays.items():
        if array.transient:
            continue

        if isinstance(array, data.Array):
            axes = field_info[name].axes

            shape = [f"__{name}_{axis}_size" for axis in axes] + [
                d for d in field_info[name].data_dims
            ]

            wrapper_sdfg.add_array(
                name,
                dtype=array.dtype,
                strides=array.strides,
                shape=shape,
                storage=array.storage,
            )

            # Calculate memlet ranges taking the origin into account
            ranges = []
            origin = origins.get(name, origins.get("_all_", None))

            if origin is None:
                raise ValueError("Freeze origin/domain: Unspecified origin for field `{name}`.")

            # Read boundaries for axis-bound fields
            index = 0
            for cartesian_index, axis in enumerate(CartesianSpace.names):
                if axis not in axes:
                    continue
                o = origin[index]
                e = field_info[name].boundary.lower_indices[cartesian_index]
                s = inner_sdfg.arrays[name].shape[index]
                ranges.append(
                    # s - 1 because ranges are inclusive
                    (o - max(0, e), o - max(0, e) + s - 1, 1)
                )
                index += 1

            # Add data dimensions to the range
            # NOTE: origin and boundary of data dimensions are always 0
            ranges += [
                (0, d - 1, 1)  # d - 1 because ranges are inclusive
                for d in field_info[name].data_dims
            ]

            if name in inputs:
                state.add_edge(
                    state.add_read(
                        name,
                        DebugInfo(123456),  # fake DebugInfo to avoid calls to `inspect`
                    ),
                    None,
                    nsdfg,
                    name,
                    Memlet(name, subset=subsets.Range(ranges)),
                )
            if name in outputs:
                state.add_edge(
                    nsdfg,
                    name,
                    state.add_write(
                        name,
                        DebugInfo(123456),  # fake DebugInfo to avoid calls to `inspect`
                    ),
                    None,
                    Memlet(name, subset=subsets.Range(ranges)),
                )
        elif isinstance(array, data.Scalar):
            wrapper_sdfg.add_scalar(
                name,
                dtype=array.dtype,
                storage=array.storage,
                lifetime=array.lifetime,
            )
            if name in inputs:
                state.add_edge(
                    state.add_read(
                        name,
                        DebugInfo(123456),  # fake DebugInfo to avoid calls to `inspect`
                    ),
                    None,
                    nsdfg,
                    name,
                    Memlet(name),
                )
            if name in outputs:
                state.add_edge(
                    nsdfg,
                    name,
                    state.add_write(
                        name,
                        DebugInfo(123456),  # fake DebugInfo to avoid calls to `inspect`
                    ),
                    None,
                    Memlet(name),
                )


def _sdfg_specialize_symbols(wrapper_sdfg: SDFG, domain: tuple[int, ...]) -> None:
    ival, jval, kval = domain[0], domain[1], domain[2]
    for sdfg in wrapper_sdfg.all_sdfgs_recursive():
        if sdfg.parent_nsdfg_node is not None:
            symmap = sdfg.parent_nsdfg_node.symbol_mapping

            if "__I" in symmap:
                ival = symmap["__I"]
                del symmap["__I"]
            if "__J" in symmap:
                jval = symmap["__J"]
                del symmap["__J"]
            if "__K" in symmap:
                kval = symmap["__K"]
                del symmap["__K"]

        sdfg.replace_dict({"__I": ival, "__J": jval, "__K": kval})
        if "__I" in sdfg.symbols:
            sdfg.remove_symbol("__I")
        if "__J" in sdfg.symbols:
            sdfg.remove_symbol("__J")
        if "__K" in sdfg.symbols:
            sdfg.remove_symbol("__K")

        for val in ival, jval, kval:
            sym = symbolic.pystr_to_symbolic(val)
            for fsym in sym.free_symbols:
                if sdfg.parent_nsdfg_node is not None:
                    sdfg.parent_nsdfg_node.symbol_mapping[str(fsym)] = fsym
                if str(fsym) not in sdfg.symbols:
                    if str(fsym) in sdfg.parent_sdfg.symbols:
                        sdfg.add_symbol(str(fsym), stype=sdfg.parent_sdfg.symbols[str(fsym)])
                    else:
                        sdfg.add_symbol(str(fsym), stype=dtypes.int32)


def freeze_origin_domain_sdfg(
    inner_sdfg_unfrozen: SDFG,
    arg_names: list[str],
    field_info: dict[str, definitions.FieldInfo],
    *,
    layout_info: layout.LayoutInfo,
    origin: dict[str, tuple[int, ...]],
    domain: tuple[int, ...],
) -> SDFG:
    """Create a new SDFG by wrapping a _copy_ of the original SDFG and freezing it's
    origin and domain.

    This wrapping is required because we do not expect any of the inner_sdfg bounds to
    have been specialized, e.g. we expect "__I/J/K" symbols to still be present. We wrap
    the call and specialize at top level, which will then be passed as a symbol to the
    inner sdfg.

    Once we move specialization of array & maps bounds upstream, this will become moot
    and can be removed, see https://github.com/GridTools/gt4py/issues/2082.

    Dev note: we need to wrap a copy to make sure we can use caching with no side effects
    in other parts of the SDFG making pipeline.

    Args:
        inner_sdfg_unfrozen: SDFG with cartesian bounds as symbols
        arg_names: names of arguments to freeze
        field_info: full info stack on arguments
        origin: tuple of offset into the memory
        domain: tuple of size for the memory written by the stencil
    """
    inner_sdfg = copy.deepcopy(inner_sdfg_unfrozen)

    wrapper_sdfg = SDFG("frozen_" + inner_sdfg.name)
    state = wrapper_sdfg.add_state("frozen_" + inner_sdfg.name + "_state")

    inputs = set()
    outputs = set()
    for node, parent in inner_sdfg.all_nodes_recursive():
        if not isinstance(node, nodes.AccessNode) or inner_sdfg.arrays[node.data].transient:
            continue

        if node.has_reads(parent):
            inputs.add(node.data)
        if node.has_writes(parent):
            outputs.add(node.data)

    # fake DebugInfo to avoid calls to `inspect`
    nsdfg = state.add_nested_sdfg(inner_sdfg, inputs, outputs, debuginfo=DebugInfo(123456))

    _sdfg_add_arrays_and_edges(
        field_info, wrapper_sdfg, state, inner_sdfg, nsdfg, inputs, outputs, origin
    )

    # in special case of empty domain, remove entire SDFG.
    if any(d == 0 for d in domain):
        states = wrapper_sdfg.states()
        assert len(states) == 1
        for node in states[0].nodes():
            state.remove_node(node)

    # make sure that symbols are passed through to inner sdfg
    for symbol in nsdfg.sdfg.free_symbols:
        if symbol not in wrapper_sdfg.symbols:
            wrapper_sdfg.add_symbol(symbol, nsdfg.sdfg.symbols[symbol])

    # Try to inline wrapped SDFG before symbols are specialized to avoid extra views
    inline_sdfgs(wrapper_sdfg)

    _sdfg_specialize_symbols(wrapper_sdfg, domain)
    _specialize_transient_strides(wrapper_sdfg, layout_info)

    for _, _, array in wrapper_sdfg.arrays_recursive():
        if array.transient:
            array.lifetime = dtypes.AllocationLifetime.SDFG

    wrapper_sdfg.arg_names = arg_names

    return wrapper_sdfg


class SDFGManager:
    # Cache loaded SDFGs across all instances (unless caching strategy is "nocaching")
    _loaded_sdfgs: ClassVar[dict[str | pathlib.Path, SDFG]] = dict()

    def __init__(self, builder: StencilBuilder, debug_stree: bool = False) -> None:
        """
        Initializes the SDFGManager.

        Args:
          builder: The StencilBuilder instance, used for build options and caching strategy.
          debug_stree: If true, saves a string representation of the schedule tree next to the cached SDFG.
        """
        self.builder = builder
        self.debug_stree = debug_stree

    def schedule_tree(self) -> tn.ScheduleTreeRoot:
        """
        Schedule tree representation of the gtir (taken from the builder).

        This function is a three-step process:

        oir = gtir_to_oir(self.builder.gtir)
        tree_ir = oir_to_tree_ir(oir)
        schedule_tree = tree_ir_to_schedule_tree(tree_ir)
        """

        oir = GTIRToOIR().visit(self.builder.gtir)

        # Deactivate caches. We need to extend the skip list in case users have
        # specified skip as well AND we need to copy in order to not trash the
        # cache hash!
        oir_pipeline: DefaultPipeline = self.builder.options.backend_opts.get(
            "oir_pipeline", DefaultPipeline()
        )
        oir_pipeline = copy.deepcopy(oir_pipeline)
        oir_pipeline.skip.extend(
            [
                caches.IJCacheDetection,
                caches.KCacheDetection,
                caches.PruneKCacheFills,
                caches.PruneKCacheFlushes,
            ]
        )
        oir = oir_pipeline.run(oir)

        tir = OIRToTreeIR(self.builder).visit(oir)

        return TreeIRToScheduleTree().visit(tir)

    @staticmethod
    def _strip_history(sdfg: SDFG) -> None:
        # strip history from SDFG for faster save/load
        for tmp_sdfg in sdfg.all_sdfgs_recursive():
            tmp_sdfg.transformation_hist = []
            tmp_sdfg.orig_sdfg = None

    @staticmethod
    def _save_sdfg(sdfg: SDFG, path: pathlib.Path, validate: bool = False) -> None:
        if validate:
            sdfg.validate()
        SDFGManager._strip_history(sdfg)
        sdfg.save(str(path), compress=True)

    def sdfg_via_schedule_tree(self, *, validate: bool = False, simplify: bool = True) -> SDFG:
        """Lower OIR into an SDFG via Schedule Tree transpile first.

        Cache the SDFG into the manager for re-use, unless the builder has a no-caching policy.

        Args:
            validate: Validate resulting SDFG
            simplify: Simplify resulting SDFG
        """
        filename = f"{self.builder.module_name}.sdfgz"
        path = (
            pathlib.Path(os.path.relpath(self.builder.module_path.parent, pathlib.Path.cwd()))
            / filename
        )

        do_cache = self.builder.caching.name != "nocaching"
        if do_cache:
            if path in SDFGManager._loaded_sdfgs:
                return SDFGManager._loaded_sdfgs[path]
            if path.exists():
                sdfg = SDFG.from_file(path)
                SDFGManager._loaded_sdfgs[path] = sdfg
                return sdfg

        # Get the schedule tree:
        #  - we expect all 3-loops to be singled maps/for
        #  - we expect the layout to be K-JI
        #  - we expect all non-cartesian control flow to be innermost
        stree = self.schedule_tree()

        # Re-order cartesian loops to match loops with memory layout
        # - layout is _always_ given I-J-K
        # - layout is reversed, inner is higher numbers, e.g K-J-I is I=2, J=1, K=0
        layout = self.builder.backend.storage_info["layout_map"](("I", "J", "K"))
        if (layout[2] < layout[1] and layout[2] > layout[0]) or (
            layout[2] > layout[1] and layout[2] < layout[0]
        ):
            raise NotImplementedError(
                f"Layout (IJK:{layout}) is not implemented for dace:X backends"
            )
        if layout[0] < layout[1]:
            flipper = passes.SwapHorizontalMaps()
            flipper.visit(stree)

        if layout[2] != 0:
            flipper = passes.PushVerticalMapDown()
            flipper.visit(stree)

        # Create SDFG
        sdfg = stree.as_sdfg(
            validate=validate,
            simplify=simplify,
            skip={"ScalarToSymbolPromotion", "ControlFlowRaising"},
        )

        if do_cache:
            self._save_sdfg(sdfg, path)
            SDFGManager._loaded_sdfgs[path] = sdfg

            if self.debug_stree:
                stree_path = path.with_suffix(".stree.txt")
                with open(stree_path, "w+") as file:
                    file.write(stree.as_string(-1))

        return sdfg

    def _frozen_sdfg(self, *, origin: dict[str, tuple[int, ...]], domain: tuple[int, ...]) -> SDFG:
        basename = self.builder.module_path.with_suffix("")
        path = pathlib.Path(f"{basename}_{shash(origin, domain)}.sdfgz")

        # check if the same sdfg is already loaded
        do_cache = self.builder.caching.name != "nocache"
        if do_cache:
            if path in SDFGManager._loaded_sdfgs:
                return SDFGManager._loaded_sdfgs[path]
            if path.exists():
                sdfg = SDFG.from_file(path)
                SDFGManager._loaded_sdfgs[path] = sdfg
                return sdfg

        # Otherwise, wrap and save sdfg from scratch
        sdfg = self.sdfg_via_schedule_tree()
        frozen_sdfg = freeze_origin_domain_sdfg(
            sdfg,
            arg_names=[arg.name for arg in self.builder.gtir.api_signature],
            field_info=make_args_data_from_gtir(self.builder.gtir_pipeline).field_info,
            layout_info=self.builder.backend.storage_info,
            origin=origin,
            domain=domain,
        )

        if do_cache:
            SDFGManager._loaded_sdfgs[path] = frozen_sdfg
            self._save_sdfg(frozen_sdfg, path)

        return frozen_sdfg

    def frozen_sdfg(self, *, origin: dict[str, tuple[int, ...]], domain: tuple[int, ...]) -> SDFG:
        return copy.deepcopy(self._frozen_sdfg(origin=origin, domain=domain))


class DaCeExtGenerator(BackendCodegen):
    def __init__(self, class_name: str, module_name: str, backend: BaseDaceBackend) -> None:
        self.class_name = class_name
        self.module_name = module_name
        self.backend = backend

    def __call__(self) -> dict[str, dict[str, str]]:
        manager = SDFGManager(self.backend.builder)

        sdfg = manager.sdfg_via_schedule_tree()
        _specialize_transient_strides(
            sdfg,
            self.backend.storage_info,
        )
        sdfg.simplify(validate=True, skip={"ScalarToSymbolPromotion"})

        # NOTE
        # The glue code in DaCeComputationCodegen.apply() (just below) will define all the
        # symbols. Our job creating the sdfg/stree is to make sure we use the same symbols
        # and to be sure that these symbols are added as dace symbols.

        implementation = DaCeComputationCodegen.apply(self.backend.builder, sdfg)

        bindings = DaCeBindingsCodegen.apply(sdfg, self.module_name, backend=self.backend)

        bindings_ext = "cu" if self.backend.storage_info["device"] == "gpu" else "cpp"
        return {
            "computation": {"computation.hpp": implementation},
            "bindings": {f"bindings.{bindings_ext}": bindings},
        }


class DaCeComputationCodegen:
    template = as_mako(
        """\
auto ${name}(const std::array<gt::uint_t, 3>& domain) {
    return [domain](${",".join(functor_args)}) {
        const int __I = domain[0];
        const int __J = domain[1];
        const int __K = domain[2];
        ${name}${state_suffix} dace_handle;
        ${backend_specifics}
        auto allocator = gt::sid::cached_allocator(&${allocator}<char[]>);
        ${"\\n".join(tmp_allocs)}
        __program_${name}(${",".join(["&dace_handle", *dace_args])});
    };
}
"""
    )

    def generate_tmp_allocs(self, sdfg: SDFG) -> list[str]:
        global_fmt = (
            "__{sdfg_id}_{name} = allocate(allocator, gt::meta::lazy::id<{dtype}>(), {size})();"
        )
        threadlocal_fmt = (
            "__{sdfg_id}_{name} = __full__{sdfg_id}_{name} + omp_get_thread_num() * ({local_size});"
        )
        res = []
        for array_sdfg, name, array in sdfg.arrays_recursive():
            if array.transient and array.lifetime == dtypes.AllocationLifetime.Persistent:
                if array.storage != dtypes.StorageType.CPU_ThreadLocal:
                    fmt = "dace_handle." + global_fmt
                    res.append(
                        fmt.format(
                            name=name,
                            sdfg_id=array_sdfg.sdfg_id,
                            dtype=array.dtype.ctype,
                            size=array.total_size,
                        )
                    )
                else:
                    fmts = [
                        "{dtype} *__full" + global_fmt,
                        "#pragma omp parallel",
                        "{{",
                        threadlocal_fmt,
                        "}}",
                    ]
                    res.extend(
                        [
                            fmt.format(
                                name=name,
                                sdfg_id=array_sdfg.sdfg_id,
                                dtype=array.dtype.ctype,
                                size=f"omp_max_threads * ({array.total_size})",
                                local_size=array.total_size,
                            )
                            for fmt in fmts
                        ]
                    )
        return res

    @staticmethod
    def _postprocess_dace_code(code_objects: list[codeobject.CodeObject], is_gpu: bool) -> str:
        lines = code_objects[[co.title for co in code_objects].index("Frame")].clean_code.split(
            "\n"
        )

        if is_gpu:
            regex = re.compile("struct [a-zA-Z_][a-zA-Z0-9_]*_t {")
            for i, line in enumerate(lines):
                if regex.match(line.strip()):
                    j = i + 1
                    while "};" not in lines[j].strip():
                        j += 1
                    lines = lines[0:i] + lines[j + 1 :]
                    break
            for i, line in enumerate(lines):
                if "#include <dace/dace.h>" in line:
                    cuda_code = next(co.clean_code for co in code_objects if co.title == "CUDA")
                    lines = lines[0:i] + cuda_code.split("\n") + lines[i + 1 :]
                    break

        def keep_line(line: str) -> bool:
            line = line.strip()
            if line == '#include "../../include/hash.h"':
                return False
            if line.startswith("DACE_EXPORTED") and line.endswith(");"):
                return False
            if line == "#include <cuda_runtime.h>":
                return False
            return True

        return "\n".join(filter(keep_line, lines))

    @classmethod
    def apply(cls, builder: StencilBuilder, sdfg: SDFG) -> str:
        self = cls()
        with config.temporary_config():
            # To prevent conflict with 3rd party usage of DaCe config always make sure that any
            #  changes be under the temporary_config manager
            if core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.ROCM:
                config.Config.set("compiler", "cuda", "backend", value="hip")
            config.Config.set("compiler", "cuda", "max_concurrent_streams", value=-1)
            config.Config.set(
                "compiler",
                "cuda",
                "default_block_size",
                value=gt_config.DACE_DEFAULT_BLOCK_SIZE,
            )
            config.Config.set("compiler", "cpu", "openmp_sections", value=False)
            code_objects = sdfg.generate_code()
        is_gpu = "CUDA" in {co.title for co in code_objects}

        has_all_sequential_compute = all(
            me.schedule == dtypes.ScheduleType.Sequential
            for me, state in sdfg.all_nodes_recursive()
            if isinstance(me, nodes.MapEntry)
        )

        if (
            not is_gpu
            and not has_all_sequential_compute
            and any(
                array.transient and array.lifetime == dtypes.AllocationLifetime.Persistent
                for *_, array in sdfg.arrays_recursive()
            )
        ):
            omp_threads = "int omp_max_threads = omp_get_max_threads();"
            omp_header = "#include <omp.h>"
        else:
            omp_threads = ""
            omp_header = ""

        interface = cls.template.definition.render(
            name=sdfg.name,
            backend_specifics=omp_threads,
            dace_args=self.generate_dace_args(builder.gtir, sdfg),
            functor_args=self.generate_functor_args(sdfg),
            tmp_allocs=self.generate_tmp_allocs(sdfg),
            allocator="gt::cuda_util::cuda_malloc" if is_gpu else "std::make_unique",
            state_suffix=config.Config.get("compiler.codegen_state_struct_suffix"),
        )
        computations = cls._postprocess_dace_code(code_objects, is_gpu)
        generated_code = f"""\
#include <gridtools/sid/sid_shift_origin.hpp>
#include <gridtools/sid/allocator.hpp>
#include <gridtools/stencil/cartesian.hpp>
{"#include <gridtools/common/cuda_util.hpp>" if is_gpu else omp_header}
namespace gt = gridtools;

{computations}

{interface}
"""

        if builder.options.format_source:
            generated_code = codegen.format_source("cpp", generated_code, style="LLVM")

        return generated_code

    def generate_dace_args(self, stencil_ir: gtir.Stencil, sdfg: SDFG) -> list[str]:
        oir = GTIRToOIR().visit(stencil_ir)
        field_extents = compute_fields_extents(oir, add_k=True)

        k_origins = {
            field_name: max(boundary[0], 0)
            for field_name, boundary in compute_k_boundary(stencil_ir).items()
        }
        offset_dict: dict[str, tuple[int, int, int]] = {
            k: (
                max(-v[0][0], 0),
                max(-v[1][0], 0),
                k_origins[k] if k in k_origins else 0,
            )
            for k, v in field_extents.items()
        }

        symbols = {f"__{var}": f"__{var}" for var in "IJK"}
        for name, array in sdfg.arrays.items():
            # transients are set to expressions based on their shape in _specialize_transient_strides
            if array.transient:
                continue

            if isinstance(array, data.Scalar):
                # will be passed by name (as variable) by the catch all below
                continue

            dims = [dim for dim, select in zip("IJK", array_dimensions(array)) if select]
            data_ndim = len(array.shape) - len(dims)

            # api field strides
            fmt = "gt::sid::get_stride<{dim}>(gt::sid::get_strides(__{name}_sid))"

            symbols.update(
                {
                    f"__{name}_{dim}_stride": fmt.format(
                        dim=f"gt::stencil::dim::{dim.lower()}", name=name
                    )
                    for dim in dims
                }
            )
            symbols.update(
                {
                    f"__{name}_d{dim}_stride": fmt.format(
                        dim=f"gt::integral_constant<int, {3 + dim}>", name=name
                    )
                    for dim in range(data_ndim)
                }
            )

            # api field pointers
            fmt = """gt::sid::multi_shifted(
                         gt::sid::get_origin(__{name}_sid)(),
                         gt::sid::get_strides(__{name}_sid),
                         std::array<gt::int_t, {ndim}>{{{origin}}}
                     )"""
            origin = tuple(
                -offset_dict[name][idx]
                for idx, var in enumerate("IJK")
                if any(
                    symbolic.pystr_to_symbolic(f"__{var}") in s.free_symbols
                    for s in array.shape
                    if hasattr(s, "free_symbols")
                )
            )
            symbols[name] = fmt.format(
                name=name,
                ndim=len(array.shape),
                origin=",".join(str(o) for o in origin),
            )

        # the remaining arguments are variables and can be passed by name
        for sym in sdfg.signature_arglist(with_types=False, for_call=True):
            if sym not in symbols:
                symbols[sym] = sym

        # return strings in order of sdfg signature
        return [symbols[s] for s in sdfg.signature_arglist(with_types=False, for_call=True)]

    def generate_functor_args(self, sdfg: SDFG) -> list[str]:
        arguments = []
        for name, array in sdfg.arrays.items():
            if array.transient:
                continue
            if isinstance(array, data.Scalar):
                arguments.append(f"auto {name}")
                continue
            if isinstance(array, data.Array):
                arguments.append(f"auto && __{name}_sid")
                continue
            raise NotImplementedError(f"generate_functor_args(): unexpected type {type(array)}")
        for name, dtype in ((n, d) for n, d in sdfg.symbols.items() if not n.startswith("__")):
            arguments.append(dtype.as_arg(name))
        return arguments


class DaCeBindingsCodegen:
    def __init__(self, backend: BaseDaceBackend) -> None:
        self.backend = backend
        self._unique_index: int = 0

    def unique_index(self) -> int:
        self._unique_index += 1
        return self._unique_index

    mako_template = bindings_main_template()

    def generate_entry_params(self, sdfg: SDFG) -> list[str]:
        res: dict[str, str] = {}

        for name in sdfg.signature_arglist(with_types=False, for_call=True):
            if name in sdfg.arrays:
                container = sdfg.arrays[name]
                if isinstance(container, data.Scalar):
                    res[name] = f"{container.ctype} {name}"
                elif isinstance(container, data.Array):
                    res[name] = (
                        "py::{pybind_type} {name}, std::array<gt::int_t,{ndim}> {name}_origin".format(
                            pybind_type=(
                                "object"
                                if self.backend.storage_info["device"] == "gpu"
                                else "buffer"
                            ),
                            name=name,
                            ndim=len(container.shape),
                        )
                    )
                else:
                    raise NotImplementedError(
                        f"generate_entry_params(): unexpected type {type(container)}"
                    )
            elif name in sdfg.symbols and not name.startswith("__"):
                res[name] = f"{sdfg.symbols[name].ctype} {name}"
        return list(res[node.name] for node in self.backend.builder.gtir.params if node.name in res)

    def generate_sid_params(self, sdfg: SDFG) -> list[str]:
        res: list[str] = []

        for name, array in sdfg.arrays.items():
            if array.transient:
                continue

            if isinstance(array, data.Scalar):
                res.append(name)
                continue

            if not isinstance(array, data.Array):
                raise NotImplementedError(f"generate_sid_params(): unexpected type {type(array)}")

            domain_dim_flags = tuple(array_dimensions(array))
            if len(domain_dim_flags) != 3:
                raise RuntimeError("Expected 3 cartesian array dimensions. Codegen error.")
            data_ndim = len(array.shape) - sum(domain_dim_flags)
            sid_def = pybuffer_to_sid(
                name=name,
                ctype=array.dtype.ctype,
                domain_dim_flags=domain_dim_flags,
                data_ndim=data_ndim,
                stride_kind_index=self.unique_index(),
                backend=self.backend,
            )

            res.append(sid_def)
        # pass scalar parameters as variables
        for name in (n for n in sdfg.symbols.keys() if not n.startswith("__")):
            res.append(name)
        return res

    def generate_sdfg_bindings(self, sdfg: SDFG, module_name: str) -> str:
        return self.mako_template.render_values(
            name=sdfg.name,
            module_name=module_name,
            entry_params=self.generate_entry_params(sdfg),
            sid_params=self.generate_sid_params(sdfg),
        )

    @classmethod
    def apply(cls, sdfg: SDFG, module_name: str, *, backend: BaseDaceBackend) -> str:
        generated_code = cls(backend).generate_sdfg_bindings(sdfg, module_name)
        if backend.builder.options.format_source:
            generated_code = codegen.format_source("cpp", generated_code, style="LLVM")
        return generated_code


class DaCePyExtModuleGenerator(PyExtModuleGenerator):
    def __init__(self, builder: StencilBuilder) -> None:
        super().__init__(builder)

    def generate_imports(self) -> str:
        return "\n".join(
            [
                *super().generate_imports().splitlines(),
                "import dace",
                "import copy",
                "from gt4py.cartesian.backend.dace_stencil_object import DaCeStencilObject",
            ]
        )

    def generate_base_class_name(self) -> str:
        return "DaCeStencilObject"

    def generate_class_members(self) -> str:
        res = super().generate_class_members()
        filepath = self.builder.module_path.joinpath(
            os.path.dirname(self.builder.module_path),
            self.builder.module_name + ".sdfgz",
        )
        res += f'\nSDFG_PATH = "{filepath}"\n'
        return res


class DaCeCUDAPyExtModuleGenerator(DaCePyExtModuleGenerator, CUDAPyExtModuleGenerator):
    pass


class BaseDaceBackend(BaseGTBackend):
    GT_BACKEND_T = "dace"
    PYEXT_GENERATOR_CLASS = DaCeExtGenerator

    def generate(self) -> type[StencilObject]:
        self.check_options(self.builder.options)

        # TODO(havogt) add bypass if computation has no effect
        self.generate_extension()

        # Generate and return the Python wrapper class
        return self.make_module()


@register
class DaceCPUBackend(BaseDaceBackend):
    name = "dace:cpu"
    languages: ClassVar[dict] = {"computation": "c++", "bindings": ["python"]}
    storage_info: ClassVar[layout.LayoutInfo] = {
        "alignment": 1,
        "device": "cpu",
        "layout_map": layout.layout_maker_factory((1, 2, 0)),  # Optimal loop order: K-I-J
        "is_optimal_layout": layout.layout_checker_factory(layout.layout_maker_factory((1, 2, 0))),
    }
    MODULE_GENERATOR_CLASS = DaCePyExtModuleGenerator

    options = BaseGTBackend.GT_BACKEND_OPTS

    def generate_extension(self) -> None:
        return self.make_extension(uses_cuda=False)


@register
class DaceCPUKFirstBackend(BaseDaceBackend):
    name = "dace:cpu_kfirst"
    languages: ClassVar[dict] = {"computation": "c++", "bindings": ["python"]}
    storage_info: ClassVar[layout.LayoutInfo] = {
        "alignment": 1,
        "device": "cpu",
        "layout_map": layout.layout_maker_factory((0, 1, 2)),  # Optimal loop order: I-J-K
        "is_optimal_layout": layout.layout_checker_factory(layout.layout_maker_factory((0, 1, 2))),
    }
    MODULE_GENERATOR_CLASS = DaCePyExtModuleGenerator

    options = BaseGTBackend.GT_BACKEND_OPTS

    def generate_extension(self) -> None:
        return self.make_extension(uses_cuda=False)


@register
class DaceCPU_KJI(BaseDaceBackend):
    name = "dace:cpu_KJI"
    languages: ClassVar[dict] = {"computation": "c++", "bindings": ["python"]}
    storage_info: ClassVar[layout.LayoutInfo] = {
        "alignment": 1,
        "device": "cpu",
        "layout_map": layout.layout_maker_factory((2, 1, 0)),  # Optimal loop order: K-J-I
        "is_optimal_layout": layout.layout_checker_factory(layout.layout_maker_factory((2, 1, 0))),
    }
    MODULE_GENERATOR_CLASS = DaCePyExtModuleGenerator

    options = BaseGTBackend.GT_BACKEND_OPTS

    def generate_extension(self) -> None:
        return self.make_extension(uses_cuda=False)


@register
class DaceGPUBackend(BaseDaceBackend):
    """DaCe python backend using gt4py.cartesian.gtc."""

    name = "dace:gpu"
    languages: ClassVar[dict] = {"computation": "cuda", "bindings": ["python"]}
    storage_info: ClassVar[layout.LayoutInfo] = {
        "alignment": 32,
        "device": "gpu",
        "layout_map": layout.layout_maker_factory((2, 1, 0)),  # Optimal loop order: K-J-I
        "is_optimal_layout": layout.layout_checker_factory(layout.layout_maker_factory((2, 1, 0))),
    }
    MODULE_GENERATOR_CLASS = DaCeCUDAPyExtModuleGenerator
    options: ClassVar[GTBackendOptions] = {
        **BaseGTBackend.GT_BACKEND_OPTS,
        "device_sync": {"versioning": True, "type": bool},
    }

    def generate_extension(self) -> None:
        return self.make_extension(uses_cuda=True)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Utilities to translate from Python types to GT4Py type specifications."""

from __future__ import annotations

import builtins
import collections.abc
import dataclasses
import enum
import functools
import pkgutil
import sys
import types
import typing
from typing import Any, ForwardRef, Optional, TypeAlias

import numpy as np
import numpy.typing as npt

from gt4py._core import definitions as core_defs
from gt4py.eve import extended_typing as xtyping, utils as eve_utils
from gt4py.next import common, named_collections
from gt4py.next.type_system import type_info, type_specifications as ts


def get_scalar_kind(dtype: npt.DTypeLike) -> ts.ScalarKind:
    """Get the GT4Py scalar kind for a given NumPy dtype."""
    # make int & float precision platform independent.
    dt: np.dtype
    if dtype is builtins.int:
        dt = np.dtype("int64")
    elif dtype is builtins.float:
        dt = np.dtype("float64")
    else:
        try:
            dt = np.dtype(dtype)
        except TypeError as err:
            raise ValueError(f"Invalid scalar type definition ('{dtype}').") from err

    if dt.hasobject:
        raise ValueError("Object dtypes are not supported.")

    if dt.shape == () and dt.fields is None:
        match dt:
            case np.bool_:
                return ts.ScalarKind.BOOL
            case np.str_:
                return ts.ScalarKind.STRING
            case np.dtype():
                return getattr(ts.ScalarKind, dt.name.upper())
            case _:
                raise ValueError(f"Impossible to map '{dtype}' value to a 'ScalarKind'.")
    else:
        raise ValueError(f"Non-trivial dtypes like '{dtype}' are not yet supported.")


def make_constructor_type(type_spec: ts.TypeSpec) -> ts.ConstructorType:
    """Create a constructor type spec for a given scalar or python type."""

    match type_spec:
        case ts.ScalarType():
            return ts.ConstructorType(
                definition=ts.FunctionType(
                    pos_only_args=[ts.DeferredType(constraint=ts.ScalarType)],
                    pos_or_kw_args={},
                    kw_only_args={},
                    returns=type_spec,
                )
            )

        case ts.NamedCollectionType() as named_tuple_type:
            type_ = pkgutil.resolve_name(named_tuple_type.original_python_type)
            pos_or_kw_args = {k: t for k, t in zip(type_spec.keys, type_spec.types)}
            kw_only_args = (
                {f.name: pos_or_kw_args.pop(f.name) for f in dataclasses.fields(type_) if f.kw_only}
                if issubclass(type_, xtyping.DataclassABC)
                else {}
            )

            return ts.ConstructorType(
                definition=ts.FunctionType(
                    pos_only_args=[],
                    pos_or_kw_args=pos_or_kw_args,
                    kw_only_args=kw_only_args,
                    returns=type_spec,
                )
            )

    raise ValueError(f"ConstructorType not implemented for {type_spec}.")


def make_type(type_: type) -> ts.TypeSpec:
    """Create a type specification for a given type."""

    if issubclass(type_, (*core_defs.SCALAR_TYPES, str)):
        return ts.ScalarType(kind=get_scalar_kind(type_))

    if issubclass(type_, named_collections.CUSTOM_NAMED_COLLECTION_TYPES):
        if issubclass(type_, xtyping.DataclassABC) and not issubclass(
            type_, named_collections.CustomDataclassNamedCollectionABC
        ):
            raise ValueError(
                f"Dataclass {type_} is not a valid field container. Supported dataclasses"
                " must have at least one member, without custom '__init__'"
                " functions, 'InitVar's or default arguments."
            )

        keys = [*named_collections.elements_keys(type_)]
        hints = named_collections.elements_types(
            type_, globalns=sys.modules[type_.__module__].__dict__
        )
        types = [from_type_hint(hint) for hint in hints.values()]
        return ts.NamedCollectionType(
            types=types, keys=keys, original_python_type=f"{type_.__module__}:{type_.__qualname__}"
        )

    if issubclass(type_, tuple) and type_ is not tuple:
        raise ValueError(f"Untyped tuple subclass {type_} is not a valid typed namedtuple.")

    raise ValueError(f"Type {type_} not supported")


def canonicalize_type_hint(
    type_hint: Any,
    *,
    globalns: Optional[dict[str, Any]] = None,
    localns: Optional[dict[str, Any]] = None,
) -> tuple[Any, tuple[Any, ...], tuple[Any, ...]]:
    """
    Canonicalize python type annotations as a tuple of (canonical_type, type_args, annotated_extra_args).
    """
    # Canonicalize 'ForwardRef()' annotations
    if isinstance(type_hint, str):
        type_hint = ForwardRef(type_hint)
    if isinstance(type_hint, ForwardRef):
        try:
            type_hint = xtyping.eval_forward_ref(type_hint, globalns=globalns, localns=localns)
        except Exception as error:
            raise ValueError(
                f"Type annotation '{type_hint}' has undefined forward references."
            ) from error

    # Cannonicalize 'Annotated' annotations
    extra_args = []
    if typing.get_origin(type_hint) is typing.Annotated:
        type_hint, *extra_args = typing.get_args(type_hint)
        if not isinstance(
            type_hint,
            collections.abc.Callable,  # type:ignore[arg-type] # see https://github.com/python/mypy/issues/14928
        ):
            type_hint = xtyping.eval_forward_ref(type_hint, globalns=globalns, localns=localns)

    canonical_type = typing.get_origin(type_hint) or type_hint
    args = typing.get_args(type_hint)

    return canonical_type, args, tuple(extra_args)


@eve_utils.optional_lru_cache(maxsize=None, typed=False)
def from_type_hint(
    type_hint: Any,
    *,
    globalns: Optional[dict[str, Any]] = None,
    localns: Optional[dict[str, Any]] = None,
) -> ts.TypeSpec:
    """Convert any kind of Python type hint to a GT4Py TypeSpec."""
    from_type_hint_same_ns = functools.partial(from_type_hint, globalns=globalns, localns=localns)

    canonical_type, args, extra_args = canonicalize_type_hint(
        type_hint, globalns=globalns, localns=localns
    )

    match canonical_type:
        case builtins.tuple:
            if not args:
                raise ValueError(f"Tuple annotation '{type_hint}' requires at least one argument.")
            if Ellipsis in args:
                raise ValueError(f"Unbound tuples '{type_hint}' are not allowed.")
            tuple_types = [from_type_hint_same_ns(arg) for arg in args]
            assert all(isinstance(elem, ts.DataType) for elem in tuple_types)
            return ts.TupleType(types=tuple_types)

        case common.Field:
            if (n_args := len(args)) != 2:
                raise ValueError(f"Field type requires two arguments, got {n_args}: '{type_hint}'.")
            dims: list[common.Dimension] = []
            dim_arg, dtype_arg = args
            dim_arg = (
                list(typing.get_args(dim_arg))
                if typing.get_origin(dim_arg) is common.Dims
                else dim_arg
            )
            if isinstance(dim_arg, list):
                for d in dim_arg:
                    if not isinstance(d, common.Dimension):
                        raise ValueError(f"Invalid field dimension definition '{d}'.")
                    dims.append(d)
            else:
                raise ValueError(f"Invalid field dimensions '{dim_arg}'.")

            try:
                dtype = from_type_hint_same_ns(dtype_arg)
            except ValueError as error:
                raise ValueError(
                    f"Field dtype argument must be a scalar type (got '{dtype_arg}')."
                ) from error
            if not isinstance(dtype, ts.ScalarType) or dtype.kind == ts.ScalarKind.STRING:
                raise ValueError("Field dtype argument must be a scalar type (got '{dtype}').")
            return ts.FieldType(dims=dims, dtype=dtype)

        case collections.abc.Callable:
            if not args:
                raise ValueError("Unannotated functions are not supported.")

            try:
                arg_types, return_type = args
                new_args = [from_type_hint_same_ns(arg) for arg in arg_types]
                assert all(isinstance(arg, ts.DataType) for arg in new_args)
            except Exception as error:
                raise ValueError(f"Invalid callable annotations in '{type_hint}'.") from error

            kwargs_info = [arg for arg in extra_args if isinstance(arg, xtyping.CallableKwargsInfo)]
            if len(kwargs_info) != 1:
                raise ValueError(f"Invalid callable annotations in '{type_hint}'.")
            kwargs = {
                arg: from_type_hint_same_ns(arg_type)
                for arg, arg_type in kwargs_info[0].data.items()
            }
            assert all(isinstance(val, (ts.DataType, ts.DeferredType)) for val in kwargs.values())

            returns = from_type_hint_same_ns(return_type)
            assert isinstance(returns, (ts.DataType, ts.DeferredType, ts.VoidType))

            # TODO(tehrengruber): print better error when no return type annotation is given
            return ts.FunctionType(
                pos_only_args=new_args,
                pos_or_kw_args=kwargs,
                kw_only_args={},  # TODO
                returns=returns,
            )

        case builtins.type if args:
            # This case matches 'type[Foo]' (where the 'Foo' type is stored in args[0])
            python_type = args[0]
            constructed_type_spec = from_type_hint_same_ns(python_type)
            return make_constructor_type(constructed_type_spec)

        case type():
            # This case matches 'int', 'float', 'FooNamedCollection' etc. used as annotations
            return make_type(canonical_type)

    raise ValueError(f"'{type_hint}' type is not supported.")


ConstantPythonNamespaceObject: TypeAlias = eve_utils.FrozenNamespace | enum.EnumMeta
PythonNamespaceObject: TypeAlias = ConstantPythonNamespaceObject | types.ModuleType


class NamespaceProxy(ts.TypeSpec):
    _object: PythonNamespaceObject

    def __getattr__(self, key: str) -> ts.TypeSpec:
        value = getattr(self._object, key)
        return from_value(value)

    def __deepcopy__(self, _: dict[int, Any]) -> NamespaceProxy:
        return NamespaceProxy(self._object)  # don't deep copy the module


def from_value(value: Any) -> ts.TypeSpec:
    """
    Make a typespec from a Python value.

    In principle, this function is conceptually equivalent to
    `from_type_hint(type(value))`, but it also handles some special cases
    which cannot be handled directly by `from_type_hint`
    (like `int` values, which could be either `int32` or `int64`
    depending on their value).
    """
    # TODO(tehrengruber): use protocol from gt4py.next.common when available
    #  instead of importing from the embedded implementation
    # TODO(tehrengruber): What we expect here currently is a GTCallable. Maybe
    #  we should check for the protocol in the future?
    if hasattr(value, "__gt_type__"):
        symbol_type = value.__gt_type__()
    elif isinstance(value, int) and not isinstance(value, bool):
        symbol_type = None
        for candidate_type in [
            ts.ScalarType(kind=ts.ScalarKind.INT32),
            ts.ScalarType(kind=ts.ScalarKind.INT64),
        ]:
            min_val, max_val = type_info.arithmetic_bounds(candidate_type)
            if min_val <= value <= max_val:
                symbol_type = candidate_type
                break
        if not symbol_type:
            raise ValueError(
                f"Value '{value}' is out of range to be representable as 'INT32' or 'INT64'."
            )
        return candidate_type
    elif isinstance(value, common.Dimension):
        symbol_type = ts.DimensionType(dim=value)
    elif isinstance(value, common.Field):
        dims = list(value.domain.dims)
        dtype = from_type_hint(value.dtype.scalar_type)
        assert isinstance(dtype, ts.ScalarType)
        symbol_type = ts.FieldType(dims=dims, dtype=dtype)
    elif isinstance(value, tuple) and not isinstance(value, xtyping.TypedNamedTupleABC):
        # Since the elements of the tuple might be one of the special cases
        # above, we can not resort to generic `infer_type` but need to do it
        # manually here. If we get rid of all the special cases this is
        # not needed anymore.
        # Note that we explicitly do not want to match namedtuples here,
        # since those should be handled as general custom types.
        elems = [from_value(el) for el in value]
        assert all(isinstance(elem, ts.DataType) for elem in elems)
        return ts.TupleType(types=elems)
    elif isinstance(value, PythonNamespaceObject):
        return NamespaceProxy(value)
    else:
        type_ = xtyping.infer_type(value, annotate_callable_kwargs=True)
        symbol_type = from_type_hint(type_)

    if isinstance(symbol_type, (ts.DataType, ts.CallableType, ts.OffsetType, ts.DimensionType)):
        return symbol_type
    else:
        raise ValueError(f"Impossible to map '{value}' value to a 'Symbol'.")


def as_dtype(type_: ts.ScalarType) -> core_defs.DType:
    """
    Translate a `ts.ScalarType` to a `core_defs.DType`

    >>> as_dtype(ts.ScalarType(kind=ts.ScalarKind.BOOL))  # doctest:+ELLIPSIS
    BoolDType(...)
    """
    if type_.kind == ts.ScalarKind.BOOL:
        return core_defs.BoolDType()
    elif type_.kind == ts.ScalarKind.INT32:
        return core_defs.Int32DType()
    elif type_.kind == ts.ScalarKind.INT64:
        return core_defs.Int64DType()
    elif type_.kind == ts.ScalarKind.FLOAT32:
        return core_defs.Float32DType()
    elif type_.kind == ts.ScalarKind.FLOAT64:
        return core_defs.Float64DType()
    raise ValueError(f"Scalar type '{type_}' not supported.")


def from_dtype(dtype: core_defs.DType) -> ts.ScalarType:
    """
    Translate a `core_defs.DType` to a `ts.ScalarType`

    >>> from_dtype(core_defs.BoolDType())  # doctest:+ELLIPSIS
    ScalarType(kind=...BOOL...)
    """
    if dtype == core_defs.BoolDType():
        return ts.ScalarType(kind=ts.ScalarKind.BOOL)
    elif dtype == core_defs.Int32DType():
        return ts.ScalarType(kind=ts.ScalarKind.INT32)
    elif dtype == core_defs.Int64DType():
        return ts.ScalarType(kind=ts.ScalarKind.INT64)
    elif dtype == core_defs.Float32DType():
        return ts.ScalarType(kind=ts.ScalarKind.FLOAT32)
    elif dtype == core_defs.Float64DType():
        return ts.ScalarType(kind=ts.ScalarKind.FLOAT64)
    raise ValueError(f"DType '{dtype}' not supported.")


# TODO(havogt): Could be extended to also accept `core_defs.DType`s as `type_`
def unsafe_cast_to(
    value: xtyping.MaybeNestedInTuple[core_defs.Scalar], type_: ts.TupleType | ts.ScalarType
) -> xtyping.MaybeNestedInTuple[core_defs.Scalar]:
    """
    Converts `value` to the type specified by `type_`.

    Note: This function does not check if the conversion is valid (within the GT4Py type system).
    It is assumed that the caller has already checked that the types are compatible.
    """
    if isinstance(type_, ts.ScalarType):
        return as_dtype(type_).scalar_type(value)
    else:
        assert isinstance(type_, ts.TupleType)
        assert isinstance(value, tuple)
        assert all(isinstance(t, (ts.ScalarType, ts.TupleType)) for t in type_.types)
        return tuple(unsafe_cast_to(v, t) for v, t in zip(value, type_.types, strict=True))  # type: ignore[arg-type] # checked in assert



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Tuple

from gt4py._core import definitions as core_defs
from gt4py.next import common, named_collections
from gt4py.next.ffront.fbuiltins import BuiltInFunction, FieldOffset, WhereBuiltinFunction


@BuiltInFunction
def as_offset(offset_: FieldOffset, field: common.Field, /) -> common.Connectivity:
    raise NotImplementedError()


@WhereBuiltinFunction
def concat_where(
    cond: common.Domain,
    true_field: common.Field | core_defs.ScalarT | Tuple | named_collections.CustomNamedCollection,
    false_field: common.Field | core_defs.ScalarT | Tuple | named_collections.CustomNamedCollection,
    /,
) -> common.Field | Tuple:
    """
    Concatenates two field fields based on a 1D mask.

    The resulting domain is the concatenation of the mask subdomains with the domains of the respective true or false fields.
    Empty domains at the beginning or end are ignored, but the interior must result in a consecutive domain.

    TODO(havogt): I can't get this doctest to run, even after copying the __doc__ in the decorator
    Example:
        >>> I = common.Dimension("I")
        >>> mask = common._field([True, False, True], domain={I: (0, 3)})
        >>> true_field = common._field([1, 2], domain={I: (0, 2)})
        >>> false_field = common._field([3, 4, 5], domain={I: (1, 4)})
        >>> assert concat_where(mask, true_field, false_field) == _field([1, 3], domain={I: (0, 2)})

        >>> mask = common._field([True, False, True], domain={I: (0, 3)})
        >>> true_field = common._field([1, 2, 3], domain={I: (0, 3)})
        >>> false_field = common._field(
        ...     [4], domain={I: (2, 3)}
        ... )  # error because of non-consecutive domain: missing I(1), but has I(0) and I(2) values
    """
    raise NotImplementedError()


EXPERIMENTAL_FUN_BUILTIN_NAMES = ["as_offset", "concat_where"]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import enum
import functools
import os
import platform
from dataclasses import dataclass
from typing import Literal, Tuple, Union

import numpy

from gt4py.cartesian import utils as gt_utils
from gt4py.cartesian.gtc.definitions import Boundary, CartesianSpace
from gt4py.cartesian.utils.attrib import (
    Any,
    AttributeClassLike,
    Dict as DictOf,
    attribclass,
    attribkwclass,
    attribute,
)


# Dev note: platform.architecture() returns "('64bit', 'ELF')" for example.
_ARCHITECTURE_LITERAL_PRECISION = platform.architecture()[0][:2]
"""Literal precision of the architecture; expected 64 or 32."""

LITERAL_INT_PRECISION = int(
    os.environ.get("GT4PY_LITERAL_INT_PRECISION", default=_ARCHITECTURE_LITERAL_PRECISION)
)
"""Default literal precision used for unspecific `int` types and casts."""

LITERAL_FLOAT_PRECISION = int(
    os.environ.get("GT4PY_LITERAL_FLOAT_PRECISION", default=_ARCHITECTURE_LITERAL_PRECISION)
)
"""Default literal precision used for unspecific `float` types and casts."""


@enum.unique
class AccessKind(enum.IntFlag):
    NONE = 0
    READ = 1
    WRITE = 2
    READ_WRITE = READ | WRITE

    def __str__(self):
        return self.name


@dataclass(frozen=True)
class DomainInfo:
    parallel_axes: Tuple[str, ...]
    sequential_axis: str
    min_sequential_axis_size: int
    ndim: int


@dataclass(frozen=True)
class FieldInfo:
    access: AccessKind
    boundary: Boundary
    axes: Tuple[str, ...]
    data_dims: Tuple[int, ...]
    dtype: numpy.dtype

    def __repr__(self):
        return "FieldInfo(access=AccessKind.{access}, boundary={boundary}, axes={axes}, data_dims={data_dims}, dtype={dtype})".format(
            access=self.access.name,
            boundary=repr(self.boundary),
            axes=repr(self.axes),
            data_dims=repr(self.data_dims),
            dtype=repr(self.dtype),
        )

    @functools.cached_property
    def domain_mask(self):
        return tuple(axis in self.axes for axis in CartesianSpace.names)

    @functools.cached_property
    def domain_ndim(self):
        return len(self.axes)

    @functools.cached_property
    def mask(self):
        return (*self.domain_mask, *((True,) * len(self.data_dims)))

    @functools.cached_property
    def ndim(self):
        return len(self.axes) + len(self.data_dims)


@dataclass(frozen=True)
class ParameterInfo:
    access: Union[Literal[AccessKind.NONE], Literal[AccessKind.READ]]
    dtype: numpy.dtype

    def __repr__(self):
        return "ParameterInfo(access=AccessKind.{access}, dtype={dtype})".format(
            access=self.access.name, dtype=repr(self.dtype)
        )


@attribkwclass
class BuildOptions(AttributeClassLike):
    """Build options."""

    name = attribute(of=str)
    module = attribute(of=str)
    format_source = attribute(of=bool, default=True)
    backend_opts = attribute(of=DictOf[str, Any], factory=dict)
    build_info = attribute(of=dict, optional=True)
    rebuild = attribute(of=bool, default=False)
    raise_if_not_cached = attribute(of=bool, default=False)
    cache_settings = attribute(of=DictOf[str, Any], factory=dict)
    _impl_opts = attribute(of=DictOf[str, Any], factory=dict)
    literal_int_precision = attribute(of=int, default=LITERAL_INT_PRECISION)
    "Literal precision for `int` types and casts. Defaults to architecture precision unless overwritten by the environment variable `GT4PY_LITERAL_INT_PRECISION`."
    literal_float_precision = attribute(of=int, default=LITERAL_FLOAT_PRECISION)
    "Literal precision for `float` types and casts. Defaults to architecture precision unless overwritten by the environment variable `GT4PY_LITERAL_FLOAT_PRECISION`."

    @property
    def qualified_name(self):
        return ".".join([self.module, self.name])

    @property
    def shashed_id(self):
        return gt_utils.shashed_id(
            self.name,
            self.module,
            self.format_source,
            self.literal_int_precision,
            self.literal_float_precision,
            *tuple(sorted(self.backend_opts.items())),
        )


@attribclass(frozen=True)
class StencilID(AttributeClassLike):
    qualified_name = attribute(of=str)
    version = attribute(of=str)

    def __iter__(self):
        return iter([self.qualified_name, self.version])


class GTError(Exception):
    pass


class GTSyntaxError(GTError):
    def __init__(self, message: str, *, frontend: str) -> None:
        super().__init__(message)
        self.frontend = frontend


class GTSpecificationError(GTError):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class GTSemanticError(GTError):
    def __init__(self, message: str) -> None:
        super().__init__(message)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from enum import Enum
from typing import Any, Optional, Sequence

import dace
from dace import (
    data as dace_data,
    properties as dace_properties,
    subsets as dace_sbs,
    symbolic as dace_sym,
    transformation as dace_transformation,
)
from dace.sdfg import graph as dace_graph, nodes as dace_nodes
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


def gt_remove_copy_chain(
    sdfg: dace.SDFG,
    validate: bool = True,
    validate_all: bool = False,
    single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
) -> Optional[int]:
    """Applies the `CopyChainRemover` transformation to the SDFG.

    The transformation returns the number of removed data containers or `None`
    if nothing was done.

    Args:
        sdfg: The SDFG to process.
        validate: Perform validation after the pass has run.
        validate_all: Perform extensive validation.
        single_use_data: Which data descriptors are used only once.
            If not passed the function will run `FindSingleUseData`.
    """

    # To ensures that the `{src,dst}_subset` are properly set, run initialization.
    #  See [issue 1703](https://github.com/spcl/dace/issues/1703)
    for state in sdfg.states():
        for edge in state.edges():
            edge.data.try_initialize(sdfg, state, edge)

    if single_use_data is None:
        find_single_use_data = dace_analysis.FindSingleUseData()
        single_use_data = find_single_use_data.apply_pass(sdfg, None)

    result: int = sdfg.apply_transformations_repeated(
        CopyChainRemover(single_use_data=single_use_data),
        validate=validate,
        validate_all=validate_all,
    )
    return result if result != 0 else None


class CopyChainRemoverMode(Enum):
    """Switch for `CopyChainRemover` mode.

    The `NONE` mode is invalid, it is there for signaling that the transformation
    cannot be applied.
    The `PULL` mode corresponds to the case where `a1` is transient, single-use
    data and its full shape is copied to `a2`, thus `a1` can be replaced with `a2`.
    The `PUSH` mode corresponds to the case where `a2` is transient, single-use
    data and its full shape is copied from `a1`, thus `a2` can be replaced with `a1`.
    """

    NONE = 0
    PULL = 1
    PUSH = 2


@dace_properties.make_properties
class CopyChainRemover(dace_transformation.SingleStateTransformation):
    """Removes chain of redundant copies, mostly related to `concat_where`.

    `concat_where`, especially when nested, will build "chains" of AccessNodes,
    this transformation will remove them. It should be called repeatedly until a
    fix point is reached and should be seen as an addition to the array removal passes
    that ship with DaCe.
    The transformation will look for the pattern `(A1) -> (A2)`, i.e. a data container
    is copied into another one, at the global scope. The transformation will then
    remove `A1` and rewire the edges such that they now refer to `A2`. Another, and
    probably better way, is to consider the transformation as fusion transformation
    for AccessNodes.

    The transformation builds on ADR-18 and imposes the following additional
    requirements before it can be applied:
    - Through the merging of `A1` and `A2` no cycles are created.
    - `A1` can not be used anywhere else.
    - `A1` is a transient and must have the same dimensionality as `A2`.
    - `A1` is fully read by `A2`.

    This transformation corresponds to the `PULL` mode of `CopyChainRemoverMode`.
    In a second iteraton, the transformation was extended to support the mirror
    case (`PUSH`), where A2 is removed if the following requirements are satisfied:
    - Through the merging of `A1` and `A2` no cycles are created.
    - `A2` can not be used anywhere else.
    - `A2` is a transient and must have the same dimensionality as `A1`.
    - `A2` is fully written by `A1`.

    Notes:
        The transformation assumes that the domain inference adjusted the ranges of
        the maps such that, in case they write into a transient, the full shape of
        the transient array is written, has the same size, i.e. there is not padding,
        or data that is not written to.

    Args:
        single_use_data: List of data containers that are used only at one place.
            Will be stored internally and not updated.

    Todo:
        - Extend such that not the full array must be read.
        - Try to allow more than one connection between `A1` and `A2`.
    """

    node_a1 = dace_transformation.PatternNode(dace_nodes.AccessNode)
    node_a2 = dace_transformation.PatternNode(dace_nodes.AccessNode)

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: dict[dace.SDFG, set[str]]

    def __init__(
        self,
        *args: Any,
        single_use_data: dict[dace.SDFG, set[str]],
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._single_use_data = single_use_data

    @classmethod
    def expressions(cls) -> Any:
        return [
            dace.sdfg.utils.node_path_graph(
                cls.node_a1,
                cls.node_a2,
            )
        ]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        a1: dace_nodes.AccessNode = self.node_a1
        a2: dace_nodes.AccessNode = self.node_a2

        # We only allow that we operate on the top level scope.
        if graph.scope_dict()[a1] is not None:
            return False

        a1_desc = a1.desc(sdfg)
        a2_desc = a2.desc(sdfg)

        # This avoids that we have to modify the subsets in a fancy way.
        # TODO(phimuell): Lift this limitation.
        if len(a1_desc.shape) != len(a2_desc.shape):
            return False

        # For simplicity we assume that neither of `a1` nor `a2` are views.
        # TODO(phimuell): Implement some of the cases.
        if gtx_transformations.utils.is_view(a1_desc, None):
            return False
        if gtx_transformations.utils.is_view(a2_desc, None):
            return False

        # TODO(phimuell): Relax this to only prevent host-device copies.
        if a1_desc.storage != a2_desc.storage:
            return False

        # There shall only be one edge connecting `a1` and `a2`.
        #  We even strengthen this requirement by not checking for the node `a2`,
        #  but for the data.
        connecting_edges = [
            oedge
            for oedge in graph.out_edges(a1)
            if isinstance(oedge.dst, dace_nodes.AccessNode) and (oedge.dst.data == a2.data)
        ]
        if len(connecting_edges) != 1:
            return False

        # The full array `a1` is copied into `a2`. Note that it is allowed, that
        #  `a2` is bigger than `a1`, it is just important that everything that was
        #  written into `a1` is also accessed.
        connecting_edge = connecting_edges[0]
        assert connecting_edge.dst is a2
        connecting_memlet = connecting_edge.data

        # If the destination or the source subset of the connection is not fully
        #  specified, we do not apply.
        src_subset = connecting_memlet.get_src_subset(connecting_edge, graph)
        if src_subset is None:
            return False
        dst_subset = connecting_memlet.get_dst_subset(connecting_edge, graph)
        if dst_subset is None:
            return False

        if self._get_copy_chain_mode(sdfg, graph) == CopyChainRemoverMode.NONE:
            return False

        # We have to ensure that no cycle is created through the removal of one node.
        #  For this we have to ensure that there is no connection, beside the direct
        #  one between `a1` and `a2`.
        # NOTE: We only check the outgoing edges of `a1`, it is not needed to also
        #   check the incoming edges, because this will not create a cycle.
        if gtx_transformations.utils.is_reachable(
            start=[oedge.dst for oedge in graph.out_edges(a1) if oedge.dst is not a2],
            target=a2,
            state=graph,
        ):
            return False

        # NOTE: In case the node we keep is a non-transient we do not have to
        #   check if it is read or written to somewhere else in this state.
        #   The reason is that ADR18 guarantees that every access is point-wise,
        #   therefore the temporary we remove is never used as double buffer.
        return True

    def is_single_use_data(
        self,
        sdfg: dace.SDFG,
        data: str | dace_nodes.AccessNode,
    ) -> bool:
        """Checks if `data` is a single use data."""
        assert sdfg in self._single_use_data
        if isinstance(data, dace_nodes.AccessNode):
            data = data.data
        return data in self._single_use_data[sdfg]

    def apply(
        self,
        graph: dace.SDFGState | dace.SDFG,
        sdfg: dace.SDFG,
    ) -> None:
        a1: dace_nodes.AccessNode = self.node_a1
        a2: dace_nodes.AccessNode = self.node_a2
        copy_chain_mode = self._get_copy_chain_mode(sdfg, graph)

        a1_to_a2_edge: dace_graph.MultiConnectorEdge = next(
            oedge for oedge in graph.out_edges(a1) if oedge.dst is a2
        )
        a1_to_a2_memlet: dace.Memlet = a1_to_a2_edge.data
        a1_to_a2_src_subset: dace_sbs.Range = a1_to_a2_memlet.get_src_subset(a1_to_a2_edge, graph)
        a1_to_a2_dst_subset: dace_sbs.Range = a1_to_a2_memlet.get_dst_subset(a1_to_a2_edge, graph)

        # Note that it is possible that `a1` is connected to the same node multiple
        #  times, although through different edges. We have to modify the data
        #  flow there, since the offsets and the data have changed. However, we must
        #  do this only once. Note that only matching the node is not enough, a
        #  counter example would be a Map with different connector names.
        reconfigured_neighbour: set[tuple[dace_nodes.Node, Optional[str]]] = set()

        #  In this section, `old_node` refers to the node, one of `a1` or `a2`,
        #  which will be replaced by the other one. The node which is kept is
        #  referred to as `new_node`.
        new_node: dace_nodes.AccessNode
        old_node: dace_nodes.AccessNode
        new_node_offsets: Sequence[dace_sym.SymbolicType]
        if copy_chain_mode == CopyChainRemoverMode.PULL:
            old_node = a1
            new_node = a2
            # Now we compose the new subset.
            #  For the `PULL` mode, we build on the fact that we have ensured that
            #  the whole array `old_node` (`a1`) is copied into `new_node` (`a2`).
            #  Thus the destination of the original source, i.e. whatever writes
            #  into `a1`, is just offset by the beginning of the range `a1` writes
            #  into `a2`.
            #       (s1) ------[c:d]-> (A1) -[0:N]------[a:b]-> (A2)
            #       (s1) ---------[(a + c):(a + c + (d - c))]-> (A2)
            #  Thus the offset is simply given by `a`, the start index where `a1`
            #  is written into `a2`.
            #  NOTE: If we ever allow the that `a1` is not fully read, then we would
            #   have to modify this computation slightly.
            new_node_offsets = a1_to_a2_dst_subset.min_element()
        else:
            assert copy_chain_mode == CopyChainRemoverMode.PUSH
            old_node = a2
            new_node = a1
            # We compose the new subset in a similar way, but using the index where
            #  `a2` starts reading from `a1`.
            new_node_offsets = a1_to_a2_src_subset.min_element()

        # Handle the producer side of things.
        for producer_edge in list(graph.in_edges(old_node)):
            producer: dace_nodes.Node = producer_edge.src
            producer_conn = producer_edge.src_conn
            if producer is new_node:
                assert producer_edge is a1_to_a2_edge
                assert copy_chain_mode == CopyChainRemoverMode.PUSH
                continue
            new_producer_edge = gtx_transformations.utils.reroute_edge(
                is_producer_edge=True,
                current_edge=producer_edge,
                ss_offset=new_node_offsets,
                state=graph,
                sdfg=sdfg,
                old_node=old_node,
                new_node=new_node,
            )
            if (producer, producer_conn) not in reconfigured_neighbour:
                gtx_transformations.utils.reconfigure_dataflow_after_rerouting(
                    is_producer_edge=True,
                    new_edge=new_producer_edge,
                    sdfg=sdfg,
                    state=graph,
                    ss_offset=new_node_offsets,
                    old_node=old_node,
                    new_node=new_node,
                )
                reconfigured_neighbour.add((producer, producer_conn))

        # Handle the consumer side of things, as they now have to read from `a2`.
        #  It is important that the offset is still the same.
        for consumer_edge in list(graph.out_edges(old_node)):
            consumer: dace_nodes.Node = consumer_edge.dst
            consumer_conn = consumer_edge.dst_conn
            if consumer is new_node:
                assert consumer_edge is a1_to_a2_edge
                assert copy_chain_mode == CopyChainRemoverMode.PULL
                continue
            new_consumer_edge = gtx_transformations.utils.reroute_edge(
                is_producer_edge=False,
                current_edge=consumer_edge,
                ss_offset=new_node_offsets,
                state=graph,
                sdfg=sdfg,
                old_node=old_node,
                new_node=new_node,
            )
            if (consumer, consumer_conn) not in reconfigured_neighbour:
                gtx_transformations.utils.reconfigure_dataflow_after_rerouting(
                    is_producer_edge=False,
                    new_edge=new_consumer_edge,
                    sdfg=sdfg,
                    state=graph,
                    ss_offset=new_node_offsets,
                    old_node=old_node,
                    new_node=new_node,
                )
                reconfigured_neighbour.add((consumer, consumer_conn))

        # After the rerouting we have to delete the transient data node and descriptor,
        #  this will also remove all the old edges.
        graph.remove_node(old_node)
        sdfg.remove_data(old_node.data, validate=False)

        # We will now propagate the strides starting from the access node we keep.
        #  Essentially, this will replace the strides from `old_node` with the ones
        #  of `new_node`. We do it outside to make sure that we do not forget a
        #  case and that we propagate the change into every NestedSDFG only once.
        gtx_transformations.gt_propagate_strides_from_access_node(
            sdfg=sdfg,
            state=graph,
            outer_node=new_node,
        )

    def _get_copy_chain_mode(self, sdfg: dace.SDFG, state: dace.SDFG) -> CopyChainRemoverMode:
        """
        Helper function to detect whether the `CopyChainRemover` tranbsformation
        should be applied in `PULL` or `PUSH` mode, for details see `CopyChainRemoverMode`.
        """
        a1: dace_nodes.AccessNode = self.node_a1
        a2: dace_nodes.AccessNode = self.node_a2

        a1_desc = a1.desc(sdfg)
        a1_range = dace_sbs.Range.from_array(a1_desc)

        a2_desc = a2.desc(sdfg)
        a2_range = dace_sbs.Range.from_array(a2_desc)

        # There shall only be one edge connecting `a1` and `a2`.
        #  We even strengthen this requirement by not checking for the node `a2`,
        #  but for the data.
        connecting_edges = [
            oedge
            for oedge in state.out_edges(a1)
            if isinstance(oedge.dst, dace_nodes.AccessNode) and (oedge.dst.data == a2.data)
        ]
        assert len(connecting_edges) == 1

        # The full array `a1` is copied into `a2`. Note that it is allowed, that
        #  `a2` is bigger than `a1`, it is just important that everything that was
        #  written into `a1` is also accessed.
        connecting_edge = connecting_edges[0]
        assert connecting_edge.dst is a2
        connecting_memlet = connecting_edge.data

        # If the destination or the source subset of the connection is not fully
        #  specified, we do not apply.
        src_subset = connecting_memlet.get_src_subset(connecting_edge, state)
        assert src_subset is not None
        dst_subset = connecting_memlet.get_dst_subset(connecting_edge, state)
        assert dst_subset is not None

        # Checking if the whole array is read.
        # NOTE: The main benefit of requiring that the whole array is read is that we
        #   do not have to adjust maps.
        # NOTE: In previous versions there was an ad hoc rule, to bypass the "full
        #   read rule". However, it caused problems, so it was removed.
        # TODO: We have to improve this test, because sometimes the expressions are
        #   so complex that without information about relations, such as
        #   `vertical_start <= vertical_end` it is not possible to prove this check.
        if a1_desc.transient and self.is_single_use_data(sdfg, a1) and src_subset.covers(a1_range):
            return CopyChainRemoverMode.PULL

        # Checking if the whole array is written.
        if a2_desc.transient and self.is_single_use_data(sdfg, a2) and dst_subset.covers(a2_range):
            return CopyChainRemoverMode.PUSH

        return CopyChainRemoverMode.NONE


@dace_properties.make_properties
class DoubleWriteRemover(dace_transformation.SingleStateTransformation):
    """Removes Redundant writes that are related to slices.

    Consider the following code:
    ```python
    for i in range(N):
        temp[i] = ...
    A[:, slice1] = temp
    A[:, slice2] = temp
    ```
    The function will transform it into the following:
    ```python
    for i in range(N):
        temp_scalar = ...
        A[i, slice1] = temp_scalar
        A[i, slice2] = temp_scalar
    ```

    The transformation will only apply if:
    - `temp` is single use data.
    - `temp` is only consumed by other AccessNodes and not Maps.
    - `temp` must be fully copied into its destination, i.e. not partially.

    Args:
        single_use_data: List of data containers that are used only at one place.
            Will be stored internally and not updated.

    Todo:
        - Extend such that not the full array must be read.
    """

    map_exit = dace_transformation.PatternNode(dace_nodes.MapExit)
    temp_node = dace_transformation.PatternNode(dace_nodes.AccessNode)

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    def __init__(
        self,
        *args: Any,
        single_use_data: Optional[dict[dace.SDFG, set[str]]],
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._single_use_data = single_use_data

    @classmethod
    def expressions(cls) -> Any:
        return [
            dace.sdfg.utils.node_path_graph(
                cls.map_exit,
                cls.temp_node,
            )
        ]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        map_exit: dace_nodes.MapExit = self.map_exit
        temp_node: dace_nodes.AccessNode = self.temp_node
        temp_desc = temp_node.desc(sdfg)

        # Only the Map can produce data.
        if graph.in_degree(temp_node) != 1:
            return False

        producer_edge = next(iter(graph.in_edges(temp_node)))
        produced_sbs: dace_sbs.Range = producer_edge.data.get_dst_subset(producer_edge, graph)
        assert produced_sbs is not None

        # Only step 1 is allowed.
        if any(step != 1 for _, _, step in map_exit.map.range):
            return False
        if any(step != 1 for _, _, step in produced_sbs):
            return False

        # To simplify implementation we only allow that there is one producer inside the Map.
        assert producer_edge.src_conn.startswith("OUT_")
        inner_producer_edges = [
            inner_map_edge
            for inner_map_edge in graph.in_edges_by_connector(
                map_exit, "IN_" + producer_edge.src_conn[4:]
            )
        ]
        if len(inner_producer_edges) != 1:
            return False
        inner_producer_edge = inner_producer_edges[0]

        # To simplify implementation, ensures that `temp_node` does not have any "dummy"
        #  dimensions, i.e. dimensions of size 1. Or differently, that every Map parameter
        #  is used in indexing.
        # TODO(phimuell): Lift this limitation.
        # TODO(phimuell): Is this check strong enough.
        if len(map_exit.map.params) != produced_sbs.dims():
            return False

        # To further simplify, we require that the Memlet is "simple" and that only a
        #  scalar is moved around.
        inner_production_sbs = inner_producer_edge.data.get_dst_subset(inner_producer_edge, graph)
        if inner_production_sbs.num_elements() != 1:
            return False
        if inner_producer_edge.data.allow_oob:
            return False
        if inner_producer_edge.data.is_empty():
            return False
        if inner_producer_edge.data.wcr is not None:
            return False

        # We only allow arrays that are transients.
        if not isinstance(temp_desc, dace_data.Array):
            return False
        if isinstance(temp_desc, dace_data.View):
            return False
        if not temp_desc.transient:
            return False

        # Check the consumer
        for consumer_edge in graph.out_edges(temp_node):
            consumer_sbs: dace_sbs.Range = consumer_edge.data.get_src_subset(consumer_edge, graph)
            if consumer_sbs is None:
                return False

            # What is produced must be consumed.
            if consumer_sbs != produced_sbs:
                return False
            if any(step != 1 for _, _, step in consumer_sbs):
                return False

            # Since `temp_node` is eliminated, the consumer must be another AccessNode.
            # TODO(phimuell): Allow the consumer to be a Map, in which case `temp_node`
            #   should not be removed.`
            consumer_node = consumer_edge.dst
            if not isinstance(consumer_node, dace_nodes.AccessNode):
                return False

            consumer_desc = consumer_node.desc(sdfg)
            if not isinstance(consumer_desc, dace_data.Array):
                return False
            if isinstance(consumer_desc, dace_data.View):
                return False

        # `temp_node` will be removed this means it must be single use.
        if self._single_use_data is None:
            find_single_use_data = dace_analysis.FindSingleUseData()
            single_use_data = find_single_use_data.apply_pass(sdfg, None)
        else:
            single_use_data = self._single_use_data
        if temp_node.data not in single_use_data[sdfg]:
            return False

        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        map_exit: dace_nodes.MapExit = self.map_exit
        temp_node: dace_nodes.AccessNode = self.temp_node
        temp_desc = temp_node.desc(sdfg)

        # Understand how the Map writes into the `temp_node`. For this we have to
        #  look at the subset of the inner edge that writes into the node.
        assert graph.in_degree(temp_node) == 1
        outer_producer_edge = next(iter(graph.in_edges(temp_node)))
        outer_producer_subset = outer_producer_edge.data.get_dst_subset(outer_producer_edge, graph)
        assert outer_producer_subset is not None
        assert outer_producer_edge.src_conn.startswith("OUT_")
        inner_producer_edge = next(
            iter(
                inner_producer_edge
                for inner_producer_edge in graph.in_edges_by_connector(
                    map_exit, "IN_" + outer_producer_edge.src_conn[4:]
                )
            )
        )
        inner_producer_subset = inner_producer_edge.data.get_dst_subset(inner_producer_edge, graph)

        # Because we distribute to multiple consumers, we need some temporary storage
        #  inside the Map scope, aka. AccessNode. We will first check if there is
        #  already an AccessNode that can be used or if we have to create a new one.
        current_inner_producer_node = inner_producer_edge.src
        if not isinstance(current_inner_producer_node, dace_nodes.AccessNode):
            reuse_inner_producer_node = False
        else:
            current_inner_producer_desc = current_inner_producer_node.desc(sdfg)
            if isinstance(current_inner_producer_desc, dace_data.View):
                reuse_inner_producer_node = False
            elif isinstance(current_inner_producer_desc, dace_data.Scalar):
                reuse_inner_producer_node = True
            elif (
                isinstance(current_inner_producer_desc, dace_data.Array)
                and len(current_inner_producer_desc.shape) == 1
                and (current_inner_producer_desc.shape[0] == 1) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
            ):
                reuse_inner_producer_node = True
            else:
                reuse_inner_producer_node = False

        if reuse_inner_producer_node:
            # We can reuse the one that is already there.
            inner_distribution_node: dace_nodes.AccessNode = current_inner_producer_node
            assert inner_producer_edge.src_conn is None
        else:
            # We have to create a new intermediate storage inside the Map scope.
            #  For this we are using a scalar.
            assert inner_producer_edge.data.dst_subset.num_elements() == 1
            distribution_data, _ = sdfg.add_scalar(
                "__gtx_double_write_remover_inner_inner_distribution_node",
                dtype=temp_desc.dtype,
                storage=dace.dtypes.StorageType.Register,
                transient=True,
                find_new_name=True,
            )
            inner_distribution_node = graph.add_access(distribution_data)
            graph.add_edge(
                inner_producer_edge.src,
                inner_producer_edge.src_conn,
                inner_distribution_node,
                None,
                dace.Memlet(
                    data=distribution_data,
                    subset="0",
                    other_subset=inner_producer_edge.data.src_subset,
                ),
            )

        # Now we have to delete the producer Memlet tree, which consists of
        #  `{inner, outer}_producer_edge`.
        # NOTE: `outer_producer_edge` and `inner_producer_edge` are no invalid.
        graph.remove_edge(outer_producer_edge)
        map_exit.remove_out_connector(outer_producer_edge.src_conn)
        graph.remove_edge(inner_producer_edge)
        map_exit.remove_in_connector(inner_producer_edge.dst_conn)
        assert graph.in_degree(temp_node) == 0 and graph.out_degree(temp_node) > 0

        # Now reroute the dataflow, instead going through `temp_node` directly serve
        #  them from the Map. To be precise use the `inner_distribution_node`.
        for consumer_edge in list(graph.out_edges(temp_node)):
            consumer_node = consumer_edge.dst
            consumer_destination = consumer_edge.data.get_dst_subset(consumer_edge, graph)
            assert isinstance(consumer_node, dace_nodes.AccessNode)

            # Since we want to understand how the destination is written we have to
            #  pass `dst_subset` as `sbs1` and `src_subset`, which reads from
            #  `temp_node` as `sbs2`.
            consumer_to_temp_mapping, consumer_drop, temp_drop = (
                gtx_transformations.utils.associate_dimmensions(
                    sbs1=consumer_edge.data.dst_subset,
                    sbs2=consumer_edge.data.src_subset,
                )
            )
            assert len(temp_drop) == 0

            # Now compose the subset (on the inside of the Map) that is used to write
            #  into `consumer_node`.
            inner_map_output_subset: list[tuple[Any, Any, int]] = []
            consumer_dst_subset = consumer_edge.data.dst_subset
            for consumer_dim in range(consumer_dst_subset.dims()):
                if consumer_dim in consumer_to_temp_mapping:
                    # This dimension is copied from `temp_node` into `consumer_node`.
                    #  Thus we have to use same subset that was used to write the
                    #  original `temp_node` thing.
                    temp_node_dim = consumer_to_temp_mapping[consumer_dim]

                    # Compute the correction.
                    consumer_correction = consumer_destination[consumer_dim][0]
                    temp_correction = outer_producer_subset[temp_node_dim][0]
                    correction = consumer_correction - temp_correction

                    # Compose the final subset.
                    inner_producer_idx = inner_producer_subset[temp_node_dim]
                    inner_map_output_subset.append(
                        (inner_producer_idx[0] + correction, inner_producer_idx[1] + correction, 1)
                    )
                else:
                    # The dimension is not written to by a Map parameter, which means
                    #  that it is some dummy dimension, such as an offset. Thus we use
                    #  the original subset.
                    #  There is no correction needed here.
                    assert consumer_dim in consumer_drop
                    inner_map_output_subset.append(consumer_dst_subset[consumer_dim])
            new_inner_dst_subset = dace_sbs.Range(inner_map_output_subset)

            # Now we create the new connection from the `inner_distribution_node` to
            #  the final consumer. For the connection from the MapExit to the consumer
            #  we essentially recycle the Memlet, but create a new one, to ensure that
            #  it has the canonical format.
            # TODO: We can not use `graph.add_memlet_path()` because sometimes Memlet
            #   propagation fails. Here we are reusing the subsets that were already
            #   there. Which assumes that they are correct.
            connector_name = map_exit.next_connector()
            graph.add_edge(
                inner_distribution_node,
                None,
                map_exit,
                "IN_" + connector_name,
                dace.Memlet(
                    data=consumer_edge.dst.data,
                    subset=new_inner_dst_subset,
                    other_subset=dace_sbs.Range.from_string("0"),
                ),
            )
            graph.add_edge(
                map_exit,
                "OUT_" + connector_name,
                consumer_edge.dst,
                consumer_edge.dst_conn,
                dace.Memlet(
                    data=consumer_edge.dst.data,
                    subset=consumer_edge.data.dst_subset,
                    other_subset=None,
                ),
            )
            map_exit.add_scope_connectors(connector_name)
            graph.remove_edge(consumer_edge)

        # Remove `temp_node` and the data it refers to as it is no longer needed.
        assert graph.out_degree(temp_node) == 0
        graph.remove_node(temp_node)
        sdfg.remove_data(temp_node.data, validate=False)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
from typing import Generic

from gt4py._core import definitions as core_defs
from gt4py.next import custom_layout_allocators as next_allocators
from gt4py.next.ffront import (
    foast_to_gtir,
    foast_to_past,
    func_to_foast,
    func_to_past,
    past_process_args,
    past_to_itir,
    stages as ffront_stages,
)
from gt4py.next.ffront.past_passes import linters as past_linters
from gt4py.next.iterator import ir as itir
from gt4py.next.otf import arguments, definitions, stages, toolchain, workflow


def jit_to_aot_args(
    inp: arguments.JITArgs,
) -> arguments.CompileTimeArgs:
    return arguments.CompileTimeArgs.from_concrete(*inp.args, **inp.kwargs)


def adapted_jit_to_aot_args_factory() -> workflow.Workflow[
    definitions.ConcreteProgramDef[definitions.IRDefinitionT, arguments.JITArgs],
    definitions.ConcreteProgramDef[definitions.IRDefinitionT, arguments.CompileTimeArgs],
]:
    """Wrap `jit_to_aot` into a workflow adapter to fit into backend transform workflows."""
    return toolchain.ArgsOnlyAdapter(jit_to_aot_args)


@dataclasses.dataclass(frozen=True)
class Transforms(
    workflow.MultiWorkflow[
        definitions.ConcreteProgramDef[definitions.IRDefinitionT, definitions.ArgsDefinitionT],
        definitions.CompilableProgramDef,
    ]
):
    """
    Modular workflow for transformations with access to intermediates.

    The set and order of transformation steps depends on the input type.
    Thus this workflow can be applied to DSL field operator and program definitions,
    as well as their AST representations. Even to Iterator IR programs, although in that
    case it will be a no-op.

    The input to the workflow as well as each step must be a `CompilableProgram`. The arguments
    inside the `CompilableProgram` passed to the whole workflow may be concrete (`JITArgs`)
    or compile-time (`CompileTimeArgs`). The individual steps (apart from `.aotify_args`)
    require compile-time arguments. Some of the steps can work with an empty `CompileTimeArgs` instance.
    """

    aotify_args: workflow.Workflow[
        definitions.ConcreteProgramDef[definitions.IRDefinitionT, arguments.JITArgs],
        definitions.ConcreteProgramDef[definitions.IRDefinitionT, arguments.CompileTimeArgs],
    ] = dataclasses.field(default_factory=adapted_jit_to_aot_args_factory)

    func_to_foast: workflow.Workflow[
        ffront_stages.ConcreteDSLFieldOperatorDef, ffront_stages.ConcreteFOASTOperatorDef
    ] = dataclasses.field(default_factory=func_to_foast.adapted_func_to_foast_factory)

    func_to_past: workflow.Workflow[
        ffront_stages.ConcreteDSLProgramDef, ffront_stages.ConcretePASTProgramDef
    ] = dataclasses.field(default_factory=func_to_past.adapted_func_to_past_factory)

    foast_to_itir: workflow.Workflow[
        ffront_stages.ConcreteFOASTOperatorDef, itir.FunctionDefinition
    ] = dataclasses.field(default_factory=foast_to_gtir.adapted_foast_to_gtir_factory)

    field_view_op_to_prog: workflow.Workflow[
        ffront_stages.ConcreteFOASTOperatorDef, ffront_stages.ConcretePASTProgramDef
    ] = dataclasses.field(default_factory=foast_to_past.operator_to_program_factory)

    past_lint: workflow.Workflow[
        ffront_stages.ConcretePASTProgramDef, ffront_stages.ConcretePASTProgramDef
    ] = dataclasses.field(default_factory=past_linters.adapted_linter_factory)

    field_view_prog_args_transform: workflow.Workflow[
        ffront_stages.ConcretePASTProgramDef, ffront_stages.ConcretePASTProgramDef
    ] = dataclasses.field(default_factory=past_process_args.transform_program_args_factory)

    past_to_itir: workflow.Workflow[
        ffront_stages.ConcretePASTProgramDef, definitions.CompilableProgramDef
    ] = dataclasses.field(default_factory=past_to_itir.past_to_gtir_factory)

    def step_order(self, inp: definitions.ConcreteProgramDef) -> list[str]:
        steps: list[str] = []
        if isinstance(inp.args, arguments.JITArgs):
            steps.append("aotify_args")
        match inp.data:
            case ffront_stages.DSLFieldOperatorDef():
                steps.extend(
                    [
                        "func_to_foast",
                        "field_view_op_to_prog",
                        "past_lint",
                        "field_view_prog_args_transform",
                        "past_to_itir",
                    ]
                )
            case ffront_stages.FOASTOperatorDef():
                steps.extend(
                    [
                        "field_view_op_to_prog",
                        "past_lint",
                        "field_view_prog_args_transform",
                        "past_to_itir",
                    ]
                )
            case ffront_stages.DSLProgramDef():
                steps.extend(
                    [
                        "func_to_past",
                        "past_lint",
                        "field_view_prog_args_transform",
                        "past_to_itir",
                    ]
                )
            case ffront_stages.PASTProgramDef():
                steps.extend(["past_lint", "field_view_prog_args_transform", "past_to_itir"])
            case itir.Program():
                pass
            case _:
                raise ValueError("Unexpected input.")
        return steps


DEFAULT_TRANSFORMS: Transforms = Transforms()


# TODO(tehrengruber): Rename class and `executor` & `transforms` attribute. Maybe:
#  `Backend` -> `Toolchain`
#  `transforms` -> `frontend_transforms`
#  `executor` -> `backend_transforms`
@dataclasses.dataclass(frozen=True)
class Backend(Generic[core_defs.DeviceTypeT]):
    name: str
    executor: workflow.Workflow[definitions.CompilableProgramDef, stages.ExecutableProgram]
    allocator: next_allocators.FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]
    transforms: workflow.Workflow[definitions.ConcreteProgramDef, definitions.CompilableProgramDef]

    def compile(
        self, program: definitions.IRDefinitionT, compile_time_args: arguments.CompileTimeArgs
    ) -> stages.ExecutableProgram:
        return self.executor(
            self.transforms(definitions.ConcreteProgramDef(data=program, args=compile_time_args))
        )

    @property
    def __gt_allocator__(
        self,
    ) -> next_allocators.FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]:
        return self.allocator



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""
Optimizable Intermediate Representation (working title).

OIR represents a computation at the level of GridTools stages and multistages,
e.g. stage merging, staged computations to compute-on-the-fly, cache annotations, etc.
"""

from __future__ import annotations

from typing import Any, List, Optional, Tuple, Type, Union

from gt4py import eve
from gt4py.cartesian.gtc import common
from gt4py.cartesian.gtc.common import AxisBound, BaseAxisBound, LocNode, RuntimeAxisBound
from gt4py.eve import datamodels


@eve.utils.noninstantiable
class Expr(common.Expr):
    pass


@eve.utils.noninstantiable
class Stmt(common.Stmt):
    pass


class IteratorAccess(Expr):
    class AxisName(eve.StrEnum):
        I = "I"  # noqa: E741 [ambiguous-variable-name]
        J = "J"
        K = "K"

    name: AxisName


class CodeBlock(common.BlockStmt[Stmt], Stmt):
    label: str


class Literal(common.Literal, Expr):
    pass


class ScalarAccess(common.ScalarAccess, Expr):
    pass


class VariableKOffset(common.VariableKOffset[Expr]):
    pass


class AbsoluteKIndex(common.AbsoluteKIndex[Expr]):
    """See gtc.common.AbsoluteKIndex"""

    pass


class FieldAccess(common.FieldAccess[Expr, VariableKOffset], Expr):
    pass


class AssignStmt(common.AssignStmt[Union[ScalarAccess, FieldAccess], Expr], Stmt):
    @datamodels.validator("left")
    def no_horizontal_offset_in_assignment(
        self, attribute: datamodels.Attribute, value: Union[ScalarAccess, FieldAccess]
    ) -> None:
        if isinstance(value, FieldAccess):
            offsets = value.offset.to_dict()
            if offsets["i"] != 0 or offsets["j"] != 0:
                raise ValueError("Lhs of assignment must not have a horizontal offset.")

    _dtype_validation = common.assign_stmt_dtype_validation(strict=True)


class MaskStmt(Stmt):
    mask: Expr
    body: List[Stmt]

    @datamodels.validator("mask")
    def mask_is_boolean_field_expr(self, attribute: datamodels.Attribute, v: Expr) -> None:
        if v.dtype != common.DataType.BOOL:
            raise ValueError("Mask must be a boolean expression.")


class HorizontalRestriction(common.HorizontalRestriction[Stmt], Stmt):
    pass


class UnaryOp(common.UnaryOp[Expr], Expr):
    pass


class BinaryOp(common.BinaryOp[Expr], Expr):
    _dtype_propagation = common.binary_op_dtype_propagation(strict=True)


class TernaryOp(common.TernaryOp[Expr], Expr):
    _dtype_propagation = common.ternary_op_dtype_propagation(strict=True)


class Cast(common.Cast[Expr], Expr):
    pass


class NativeFuncCall(common.NativeFuncCall[Expr], Expr):
    _dtype_propagation = common.native_func_call_dtype_propagation(strict=True)


class While(common.While[Stmt, Expr], Stmt):
    pass


class Decl(LocNode):
    name: eve.Coerced[eve.SymbolName]
    dtype: common.DataType

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if type(self) is Decl:
            raise TypeError("Trying to instantiate `Decl` abstract class.")
        super().__init__(*args, **kwargs)


class FieldDecl(Decl):
    dimensions: Tuple[bool, bool, bool]
    data_dims: Tuple[int, ...] = eve.field(default_factory=tuple)


class ScalarDecl(Decl):
    pass


class LocalScalar(Decl):
    pass


class Temporary(FieldDecl):
    pass


def _check_interval(instance: Union[Interval, UnboundedInterval]) -> None:
    start, end = instance.start, instance.end
    if (
        start is not None
        and start.level == common.LevelMarker.END
        and end is not None
        and end.level == common.LevelMarker.START
    ):
        raise ValueError("Start level must be smaller or equal end level")

    if (
        start is not None
        and end is not None
        and start.level == end.level
        and not isinstance(start, RuntimeAxisBound)
        and not isinstance(end, RuntimeAxisBound)
        and not start.offset < end.offset  # type: ignore
    ):
        raise ValueError(
            "Start offset must be smaller than end offset if start and end levels are equal"
        )


class Interval(LocNode):
    start: BaseAxisBound
    end: BaseAxisBound

    @datamodels.root_validator
    @classmethod
    def check(cls: Type[Interval], instance: Interval) -> None:
        _check_interval(instance)

    def covers(self, other: Interval) -> bool:
        if not (
            isinstance(self.start, AxisBound)
            and isinstance(self.end, AxisBound)
            and isinstance(other.start, AxisBound)
            and isinstance(other.end, AxisBound)
        ):
            raise NotImplementedError("Can't determine coverage of runtime intervals")
        outer_starts_lower = self.start < other.start or self.start == other.start
        outer_ends_higher = self.end > other.end or self.end == other.end
        return outer_starts_lower and outer_ends_higher

    def intersects(self, other: Interval) -> bool:
        if not (
            isinstance(self.start, AxisBound)
            and isinstance(self.end, AxisBound)
            and isinstance(other.start, AxisBound)
            and isinstance(other.end, AxisBound)
        ):
            raise NotImplementedError("Can't determine intersection of runtime intervals")
        return not (other.start >= self.end or self.start >= other.end)

    def shifted(self, offset: Optional[int]) -> Union[Interval, UnboundedInterval]:
        if not (isinstance(self.start, AxisBound) and isinstance(self.end, AxisBound)):
            raise NotImplementedError("Can't shift runtime-intervals")
        if offset is None:
            return UnboundedInterval()
        start = AxisBound(level=self.start.level, offset=self.start.offset + offset)
        end = AxisBound(level=self.end.level, offset=self.end.offset + offset)
        return Interval(start=start, end=end)

    def has_runtime_access(self) -> bool:
        """Check if the interval contains runtime-accesses.

        Returns:
            bool: True if there are runtime bounds to the interval.
        """
        return isinstance(self.start, RuntimeAxisBound) or isinstance(self.end, RuntimeAxisBound)

    @classmethod
    def full(cls):
        return cls(start=AxisBound.start(), end=AxisBound.end())


class UnboundedInterval:
    start: Optional[AxisBound] = None
    end: Optional[AxisBound] = None

    @datamodels.root_validator
    @classmethod
    def check(cls: Type[UnboundedInterval], instance: UnboundedInterval) -> None:
        _check_interval(instance)

    def covers(self, other: Union[Interval, UnboundedInterval]) -> bool:
        if not (
            isinstance(self.start, AxisBound)
            and isinstance(self.end, AxisBound)
            and isinstance(other.start, AxisBound)
            and isinstance(other.end, AxisBound)
        ):
            raise NotImplementedError("Can't assess if runtime intervals cover each other")

        if self.start is None and self.end is None:
            return True
        if (
            self.end is None
            and other.start is not None
            and self.start is not None
            and other.start >= self.start
        ):
            return True
        if (
            self.start is None
            and other.end is not None
            and self.end is not None
            and other.end <= self.end
        ):
            return True

        # at this point, we know self is actually bounded, so can't cover unbounded intervals
        if other.start is None or other.end is None:
            return False

        assert isinstance(other, Interval)
        return Interval(start=self.start, end=self.end).covers(other)

    def intersects(self, other: Union[Interval, UnboundedInterval]) -> bool:
        if not (
            isinstance(self.start, AxisBound)
            and isinstance(self.end, AxisBound)
            and isinstance(other.start, AxisBound)
            and isinstance(other.end, AxisBound)
        ):
            raise NotImplementedError("Can't assess if runtime intervals intersect")

        no_overlap_high = (
            self.end is not None and other.start is not None and other.start >= self.end
        )
        no_overlap_low = (
            self.start is not None and other.end is not None and self.start >= other.end
        )
        return not (no_overlap_low or no_overlap_high)

    def shifted(self, offset: Optional[int]) -> UnboundedInterval:
        if offset is None:
            return UnboundedInterval()

        start = (
            None
            if self.start is None
            else AxisBound(level=self.start.level, offset=self.start.offset + offset)
        )
        end = (
            None
            if self.end is None
            else AxisBound(level=self.end.level, offset=self.end.offset + offset)
        )
        return UnboundedInterval(start=start, end=end)  # type: ignore

    @classmethod
    def full(cls):
        return cls()


class HorizontalExecution(LocNode, eve.SymbolTableTrait):
    body: List[Stmt]
    declarations: List[LocalScalar]


class CacheDesc(LocNode):
    name: eve.Coerced[eve.SymbolRef]


class IJCache(CacheDesc):
    pass


class KCache(CacheDesc):
    fill: bool
    flush: bool


class VerticalLoopSection(LocNode):
    interval: Interval
    horizontal_executions: List[HorizontalExecution]


class VerticalLoop(LocNode):
    loop_order: common.LoopOrder
    sections: List[VerticalLoopSection]
    caches: List[CacheDesc] = eve.field(default_factory=list)

    @datamodels.validator("sections")
    def nonempty_loop(self, attribute: datamodels.Attribute, v: List[VerticalLoopSection]) -> None:
        if not v:
            raise ValueError("Empty vertical loop is not allowed")

    @datamodels.root_validator
    @classmethod
    def valid_section_intervals(cls: Type[VerticalLoop], instance: VerticalLoop) -> None:
        starts, ends = zip(*((s.interval.start, s.interval.end) for s in instance.sections))
        if instance.loop_order == common.LoopOrder.BACKWARD:
            starts, ends = starts[:-1], ends[1:]
        else:
            starts, ends = starts[1:], ends[:-1]

        if not all(
            start.level == end.level and start.offset == end.offset
            for start, end in zip(starts, ends)
        ):
            raise ValueError("Loop intervals not contiguous or in wrong order")


class Stencil(LocNode, eve.ValidatedSymbolTableTrait):
    name: str
    # TODO: fix to be List[Union[ScalarDecl, FieldDecl]]
    params: List[Decl]
    vertical_loops: List[VerticalLoop]
    declarations: List[Temporary]

    _validate_dtype_is_set = common.validate_dtype_is_set()
    _validate_lvalue_dims = common.validate_lvalue_dims(VerticalLoop, FieldDecl)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import copy
import warnings
from typing import Final, Iterable, Optional, Sequence, TypeAlias

import dace
import sympy
from dace import data as dace_data, subsets as dace_sbs, symbolic as dace_sym
from dace.sdfg import graph as dace_graph, nodes as dace_nodes

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


InlineSpec: TypeAlias = tuple[
    dace_nodes.MapExit,
    dace_nodes.AccessNode,
    dace_nodes.MapEntry,
    dace_sbs.Range,
    dict[str, dace_sym.SymbolicType],
]
"""Specify how the inlining has to be performed.

The information is computed by the `find_nodes_to_inline()` function and returned as
its second return value (the first return value is the set of nodes that must be
inlined). The user should see it as an opaque structure.
"""


def inline_dataflow_into_map(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    edge: dace_graph.MultiConnectorEdge[dace.Memlet],
) -> Optional[tuple[dace_nodes.NestedSDFG, dace_nodes.AccessNode]]:
    """Tries to perform the inlining of `edge`.

    The function replaces the transmission, i.e. reading, of data through `edge`
    with an on-the-fly computation of the same data. The term inlining is used
    because the dataflow, that computes the transmitted data is embedded/inlined
    into the scope of `edge`.

    This function performs inlining in a single step or returns `None` if that
    is not possible. It is also possible to perform these two steps separately,
    if and how inlining is possible, see `find_nodes_to_inline()` and perform the
    actual inlining accordingly, see `perform_dataflow_inlining()`, separately.
    For more information please consult the documentation of these two functions.

    Args:
        sdfg: The SDFG on which we operate.
        state: The state in which `edge` is located.
        edge: The `edge` whose dataflow should be inlined.
    """
    extracted_information = find_nodes_to_inline(sdfg, state, edge)
    if extracted_information is None:
        return None

    return perform_dataflow_inlining(
        sdfg=sdfg,
        state=state,
        edge=edge,
        nodes_to_inline=extracted_information[0],
        inline_spec=extracted_information[1],
    )


def perform_dataflow_inlining(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    edge: dace_graph.MultiConnectorEdge[dace.Memlet],
    nodes_to_inline: set[dace_nodes.Node],
    inline_spec: InlineSpec,
) -> Optional[tuple[dace_nodes.NestedSDFG, dace_nodes.AccessNode]]:
    """Performs the second step, i.e. the actual inlining, of the dataflow.

    The function will use the information of `nodes_to_inline` and `inline_spec`,
    which was computed by `find_nodes_to_inline()`, to perform the inlining. The
    inlining essentially works by creating a nested SDFG inside the Map, where `edge`
    is located and replicate the dataflow into it. The function will also create an
    AccessNode that is used as output of the nested SDFG. Furthermore, the function
    will adjust all dataflow, that was going through `edge` such that it will pass
    through that AccessNode. As a final step the Memlet path of `edge` will be removed.

    Returns:
        The function returns two values. First it will return the nested SDFG that
        contains the dataflow needed for the computation. The second value is an
        AccessNode that is used as output of the nested SDFG.

    Args:
        sdfg: The SDFG on which we operate.
        state: The state in which `edge` is located.
        edge: The `edge` whose dataflow should be inlined.
        nodes_to_inline: The set of nodes that create the dataflow that is needed.
            This is the first return value of `find_nodes_to_inline()`.
        inline_spec: Auxiliary information how the inlining has to happen. This is
            the second return value of `find_nodes_to_inline()`.
    """
    (
        first_map_exit,
        intermediate_node,
        second_map_entry,
        exchange_subset,
        first_map_param_mapping,
    ) = inline_spec
    assert len(nodes_to_inline) > 0
    nsdfg, input_node_map, output_name = _populate_nested_sdfg(
        sdfg=sdfg,
        state=state,
        nodes_to_replicate=nodes_to_inline,
        first_map_exit=first_map_exit,
        exchange_subset=exchange_subset,
        intermediate_node=intermediate_node,
    )

    return _insert_nested_sdfg(
        sdfg=sdfg,
        state=state,
        edge_to_replace=edge,
        second_map_entry=second_map_entry,
        nsdfg=nsdfg,
        input_node_map=input_node_map,
        output_name=output_name,
        first_map_param_mapping=first_map_param_mapping,
        intermediate_node=intermediate_node,
    )


def find_nodes_to_inline(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    edge: dace_graph.MultiConnectorEdge[dace.Memlet],
) -> Optional[tuple[set[dace_nodes.Node], InlineSpec]]:
    """First step of dataflow inlining, computing the inline specification.

    The inline specification describes how the inlining of dataflow has to be done.
    The values computed by this function must be passed to `perform_dataflow_inlining()`
    for actually inlining dataflow.
    The function will analyze how a read through `edge` can be replaced by direct
    computations. If this is not possible then the function returns `None`.

    Currently the following restrictions (not exhaustive) are imposed on the dataflow:
    - `edge` must be an outgoing edge of a top level MapEntry.
    - The generating dataflow, i.e. the one that is inlined, must be inside a Map.
    - The intermediate between the two Maps can only have one producer, i.e. incoming
        degree of 1. Note that in GT4Py this means that there is no other AccessNode
        that refers to the same data with an incoming edge.
    - All data dependencies must be met. This means that `edge` can only consume data
        that is generated by a single iteration of the first data.

    Return:
        The function returns two values. The first is a `set` that contains all nodes
        that have to be inlined (the two Map nodes of the generating Map are not
        included). The second value are auxiliary data for the inlining in an
        unspecified format.

    Args:
        sdfg: The SDFG on which we operate.
        state: The state in which `edge` is located.
        edge: The `edge` whose dataflow should be inlined.
    """
    scope_dict = state.scope_dict()

    # Must be connected to a top level MapEntry.
    if not isinstance(edge.src, dace_nodes.MapEntry):
        return None
    second_map_entry: dace_nodes.MapEntry = edge.src
    if scope_dict[edge.src] is not None:
        return None

    # Check the Memlet.
    if edge.data.get_src_subset(edge, state) is None:
        return None
    if edge.data.wcr is not None:
        return None

    # We require that every access in the edge is either a literal digit, such as
    #  `1` but also `1:3` or that exactly one symbol that is also map parameter is
    #  used to do the index. Note that this does allow accesses such as
    #  `a[map_index:(map_index + 4)]` but not accesses such as `a[map_index:(map_index + symb)]`
    #  or `a[some_other_symbol]`. Furthermore, we require that all parameters of the
    #  second Map are used for the access (exception see below). These rule are used
    #  to ensure that one iteration of the first Map is enough to satisfy all data
    #  dependencies of the second Map.
    avail_second_map_param = {param for param in second_map_entry.map.params}
    literal_index_access_dims: list[int] = []
    for dim, (start, stop, step) in enumerate(edge.data.src_subset):
        if (step != 1) == True:  # noqa: E712 [true-false-comparison]  # SymPy comparison
            return None

        if str(start).isdigit():
            # The start index is a digit, we require that the end is also a digit.
            if not str(stop).isdigit():
                return None

            # There is an access done using a literal.
            literal_index_access_dims.append(dim)

        else:
            # The start index is a symbol. We require that it is an unused parameter
            #  and that the stop is also that parameter.
            start_symbols = {str(sym) for sym in start.free_symbols}
            if len(start_symbols) != 1:
                return None
            elif start_symbols.issubset(avail_second_map_param):
                stop_symbols = {str(sym) for sym in stop.free_symbols}
                if start_symbols != stop_symbols:
                    return None
                avail_second_map_param.remove(next(iter(start_symbols)))

    # Usually we require that all parameters of the second Map are used for the
    #  access. The main reason is to prevent some slicing, i.e. to ensure that
    #  the access is not done later. However, this prevents the second Map to read
    #  from a Map that has less dimensions. Thus as an exception we allow that case
    #  if all accesses are done using parameters, then not all of them have
    #  to be used.
    if (len(literal_index_access_dims) != 0) and (len(avail_second_map_param) != 0):
        return None

    # We restrict the destination nodes to Taklets and AccessNodes because these node
    #  types are simple to modify. However, even the Tasklet could be hard to handle,
    #  in some cases (such as `deref` Tasklets).
    # TODO(phimuell): Allow more node types, especially NestedSDFGs in certain cases.
    for leave_edge in state.memlet_tree(edge).leaves():
        if not isinstance(leave_edge.dst, (dace_nodes.AccessNode, dace_nodes.Tasklet)):
            return None

    # There must be an intermediate that is only generated by a single source.
    #  NOTE: We omit the check that the intermediate is single use data here.
    intermediate_node: dace_nodes.Node = next(
        iter(state.in_edges_by_connector(edge.src, "IN_" + edge.src_conn[4:]))
    ).src
    if not isinstance(intermediate_node, dace_nodes.AccessNode):
        return None
    if state.in_degree(intermediate_node) != 1:
        return None

    # The producer must be a Map.
    producing_edge = next(iter(state.in_edges(intermediate_node)))
    if not isinstance(producing_edge.src, dace_nodes.MapExit):
        return None
    first_map_exit: dace_nodes.MapExit = producing_edge.src

    # Dynamic Map ranges are not yet supported.
    # TODO(phimuell): Lift this restriction.
    first_map_entry: dace_nodes.MapEntry = state.entry_node(first_map_exit)
    if any(
        iedge.dst_conn is None or (not iedge.dst_conn.startswith("IN_"))
        for iedge in state.in_edges(first_map_entry)
    ):
        return None

    # NOTE: On an `OUT_` connector of a MapExit node multiple edges can converge,
    #   we can not handle that thus we have to restrict that case.
    writing_edges = [
        iedge
        for iedge in state.in_edges_by_connector(
            first_map_exit, "IN_" + producing_edge.src_conn[4:]
        )
    ]
    if len(writing_edges) != 1:
        return None

    first_map_param_mapping = _get_first_map_parameter_to_second_map_mapping(
        state=state,
        first_map=first_map_exit.map,
        write_edge=writing_edges[0],
        second_map=edge.src.map,
        read_edge=edge,
    )

    # Check the data dependencies. It is a little bit different compared to classical
    #  MapFusionVertical, where the two iterations are coupled. Here the restriction
    #  is that there exists _one_ iteration of the first Map that, generates the data
    #  that _another_ iteration of the second Map needs. This is a much weaker restriction
    #  but also harder to check. As a first approximation we look at the shape of the
    #  data that is generated and consumed.
    #  To avoid dynamic allocation we require that the exchange set has a constant size.
    consumer_subset = edge.data.src_subset
    consumer_shape = consumer_subset.size()
    producer_subset = writing_edges[0].data.get_dst_subset(writing_edges[0], state)
    producer_shape = producer_subset.size()
    for cshp, pshp in zip(consumer_shape, producer_shape, strict=True):
        if not str(pshp).isdigit():
            return None
        if (cshp <= pshp) == True:  # noqa: E712 [true-false-comparison]  # SymPy comparison
            continue
        return None

    # If a dimension of the intermediate is accessed not by a Map parameter, we require
    #  that that range is known at generation time, see above. In addition to that we
    #  require that the range starts at zero. The only reason for that is that it makes
    #  computing the correction offset much simpler.
    # TODO(phimuell): Allow non-zero start index for intermediate access.
    for literal_index_access_dim in literal_index_access_dims:
        if producer_subset[literal_index_access_dim][0] != 0:
            return None

    # Collect all the nodes that give rise to the data.
    generating_nodes = gtx_transformations.utils.find_upstream_nodes(
        start=first_map_exit,
        state=state,
        start_connector="IN_" + producing_edge.src_conn[4:],
        limit_node=state.entry_node(first_map_exit),
    )
    assert first_map_exit not in generating_nodes
    assert state.entry_node(first_map_exit) not in generating_nodes
    assert len(generating_nodes) >= 1

    # For syntactical reasons we require that all outgoing edges of the nodes in
    #  `generating_nodes` must lead to nodes that are also in that set, or go to
    #  the MapExit node (in which case the outgoing edge must be `edge`). The only
    #  exceptions are AccessNode. Technically we could also handle Tasklets, but
    #  they would require a little bit more work in `_populated_nsdfg()`.
    for generating_node in generating_nodes:
        for oedge in state.out_edges(generating_node):
            if oedge.dst is first_map_exit:
                assert oedge is writing_edges[0]
            elif oedge.dst not in generating_nodes:
                if isinstance(oedge.src, dace_nodes.AccessNode) and (
                    not gtx_transformations.utils.is_view(oedge.src, sdfg)
                ):
                    continue
                return None

    return (
        generating_nodes,
        (
            first_map_exit,
            intermediate_node,
            edge.src,
            producer_shape,
            first_map_param_mapping,
        ),
    )


def _insert_nested_sdfg(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    edge_to_replace: dace_graph.MultiConnectorEdge[dace.Memlet],
    second_map_entry: dace_nodes.MapEntry,
    nsdfg: dace.SDFG,
    input_node_map: dict[str, dace_nodes.Node],
    output_name: str,
    first_map_param_mapping: dict[str, dace_sym.SymbolicType],
    intermediate_node: dace_nodes.AccessNode,
) -> tuple[dace_nodes.NestedSDFG, dace_nodes.AccessNode]:
    """Helper function for the `perform_dataflow_inlining()` step.

    After `_populate_nested_sdfg()` has prepared the nested SDFG, this function
    is used to insert it into the second Map scope and to rewire the internal
    dataflow to actually use it. It will also remove the Memlet path of
    `edge_to_replace`.
    """

    # Find all data that is already inside Map.
    nodes_available_in_second_map: Final[dict[dace_nodes.AccessNode, str]] = {
        iedge.src: "OUT_" + iedge.dst_conn[3:]
        for iedge in state.in_edges(second_map_entry)
        # TODO(phimuell): Remove the view check once `_populate_nested_sdfg()` is able to handle them.
        if isinstance(iedge.src, dace_nodes.AccessNode)
        and (not gtx_transformations.utils.is_view(iedge.src, sdfg))
    }

    nsdfg_node = state.add_nested_sdfg(
        sdfg=nsdfg,
        inputs=input_node_map.keys(),
        outputs={output_name},
        symbol_mapping=first_map_param_mapping,
    )

    for input_name, required_node in input_node_map.items():
        if required_node in nodes_available_in_second_map:
            conn_to_use = nodes_available_in_second_map[required_node]

            # Update the range to ensure that everything is mapped into it.
            memlet_path_to_widen = state.memlet_path(
                next(iter(state.in_edges_by_connector(second_map_entry, "IN_" + conn_to_use[4:])))
            )
            for edge in memlet_path_to_widen:
                edge.data.try_initialize(
                    sdfg, state, edge
                )  # There is not `Memlet::set_src_subset()`.
                edge.data.src_subset = dace_sbs.Range.from_array(sdfg.arrays[required_node.data])
                if edge.dst is second_map_entry:
                    break
            else:
                raise RuntimeError("Unreachable code.")

        else:
            conn_to_use = second_map_entry.next_connector(try_name=input_name)
            second_map_entry.add_scope_connectors(conn_to_use)
            state.add_edge(
                required_node,
                None,
                second_map_entry,
                "IN_" + conn_to_use,
                dace.Memlet.from_array(required_node.data, required_node.desc(sdfg)),
            )
            conn_to_use = "OUT_" + conn_to_use

        # Connect to the nested SDFG
        state.add_edge(
            second_map_entry,
            conn_to_use,
            nsdfg_node,
            input_name,
            dace.Memlet.from_array(required_node.data, required_node.desc(sdfg)),
        )

    # In case there is no input use an empty edge to bind the nested SDFG into the scope.
    if len(input_node_map) == 0:
        state.add_edge(
            second_map_entry,
            None,
            nsdfg_node,
            None,
            dace.Memlet(),
        )

    # Now rewire the output.
    #  Instead of reading from the intermediate, we will now read from the output of the
    #  nested SDFG. For which we create a dedicated output.
    outer_output_name = sdfg.add_datadesc(
        output_name,
        nsdfg.arrays[output_name].clone(),
        find_new_name=True,
    )
    sdfg.arrays[outer_output_name].transient = True
    outer_output_node = state.add_access(outer_output_name)

    state.add_edge(
        nsdfg_node,
        output_name,
        outer_output_node,
        None,
        dace.Memlet.from_array(outer_output_name, sdfg.arrays[outer_output_name]),
    )

    offset = _compute_offset(edge_to_replace, second_map_entry.map.params)
    replaced_edge = state.add_edge(
        outer_output_node,
        None,
        edge_to_replace.dst,
        edge_to_replace.dst_conn,
        dace.Memlet(
            data=outer_output_name,
            subset=edge_to_replace.data.src_subset.offset_new(offset, negative=True),
            other_subset=edge_to_replace.data.get_dst_subset(edge_to_replace, state),
        ),
    )

    # Readjust all accesses.
    gtx_transformations.utils.reconfigure_dataflow_after_rerouting(
        is_producer_edge=False,
        new_edge=replaced_edge,
        ss_offset=offset,
        state=state,
        sdfg=sdfg,
        old_node=intermediate_node,
        new_node=outer_output_node,
    )

    # Now removing the edge that we just replaced. We will only replace at most the
    #  Memlet path, the actual source, i.e. the first Map, will still be there.
    #  To remove it, one should call a dead dataflow elimination.
    state.remove_memlet_path(edge_to_replace)

    return nsdfg_node, outer_output_node


def _populate_nested_sdfg(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    nodes_to_replicate: set[dace_nodes.Node],
    first_map_exit: dace_nodes.MapExit,
    exchange_subset: dace_sbs.Range,
    intermediate_node: dace_nodes.AccessNode,
) -> tuple[dace.SDFG, dict[str, dace_nodes.Node], str]:
    """Helper function for the `perform_dataflow_inlining()` step.

    This is the first step of the inlining. The function will generate an SDFG
    and replicate the dataflow described by `nodes_to_replicate`. It will also
    prepare the mapping of all external data into the nested SDFG.
    """
    nsdfg = dace.SDFG(f"_inline_fused_nested_sdfg_{sdfg.name}_{first_map_exit.map.label}")
    nstate = nsdfg.add_state(is_start_block=True)

    # Maps the name of the input to the node on the outside that supplies it.
    input_node_map: dict[str, dace_nodes.Node] = dict()

    # Replicate the nodes into the nested SDFG.
    replicated_node_map: dict[dace_nodes.Node, dace_nodes.Node] = {}
    for node_to_replicate in nodes_to_replicate:
        if (
            isinstance(node_to_replicate, dace_nodes.AccessNode)
            and node_to_replicate.data not in nsdfg.arrays
        ):
            dname = node_to_replicate.data
            ddesc = sdfg.arrays[dname]
            nsdfg.add_datadesc(dname, ddesc.clone())
        replicated_node = copy.deepcopy(node_to_replicate)
        replicated_node_map[node_to_replicate] = replicated_node
        nstate.add_node(replicated_node)

    # This is the name used to store the output.
    output_name: Optional[str] = None

    # Replicate the connections.
    first_map_entry = state.entry_node(first_map_exit)
    for node_to_replicate in nodes_to_replicate:
        replicated_node = replicated_node_map[node_to_replicate]

        # All outgoing edges, with one exception, lead to a node that also has been
        #  replicated, and we can thus simply copy it. The exception is the final
        #  output edge, which connects the node to the first MapExit node and would
        #  write into the intermediate. Instead of this we create a new output node
        #  node and connect to it.
        for oedge in state.out_edges(node_to_replicate):
            dst_node = oedge.dst

            if dst_node is first_map_exit:
                # A connection to the MapExit node is indicating that we now write to
                #  the output, which is ultimately the intermediate. But since we can
                #  not use it we have to create a replacement.
                if output_name is not None:
                    raise ValueError("Found multiple connection to the MapExit node.")
                assert dst_node not in nodes_to_replicate

                output_name, output_desc = nsdfg.add_array(
                    f"__inline_fuser_replacing_{intermediate_node.data}",
                    shape=tuple(exchange_subset),
                    dtype=intermediate_node.desc(sdfg).dtype,
                    find_new_name=True,
                    transient=False,
                )
                nstate.add_edge(
                    replicated_node_map[oedge.src],
                    oedge.src_conn,
                    nstate.add_access(output_name),
                    oedge.dst_conn,
                    memlet=dace.Memlet(
                        data=output_name,
                        subset=dace_sbs.Range.from_array(output_desc),
                        other_subset=oedge.data.get_src_subset(oedge, state),
                    ),
                )

            elif oedge.src not in nodes_to_replicate:
                # This connection leads to a node that has not been replicated into the
                #  nested SDFG. This is only allowed if `oedge.src` is an AccessNode in
                #  that case we simply do not generate this edge. All other cases
                #  are harder to handle.
                # NOTE: In some cases it would be possible to handle views.
                if (
                    not isinstance(oedge.src, dace_nodes.AccessNode)
                ) or gtx_transformations.utils.is_view(oedge.src, sdfg):
                    raise NotImplementedError(
                        "Connections to a non replicated node are only allowed if the source node is an AccessNode"
                    )

                warnings.warn(
                    "Detected computation of data that might not be needed in inline fuser.",
                    stacklevel=0,
                )

            else:
                # A normal connection to a node that is also mapped into the nested
                #  SDFG. We just replicate it.
                nstate.add_edge(
                    replicated_node_map[oedge.src],
                    oedge.src_conn,
                    replicated_node_map[oedge.dst],
                    oedge.dst_conn,
                    dace.Memlet.from_memlet(oedge.data),
                )

        # We have to look at the input connections separately. There are two types of
        #  them. The first type comes from a node that has been replicated and was
        #  handled by the loop above. The second kind are connections from the
        #  MapEntry node. Such connections will trigger the mapping of the data from
        #  outside the Map into the nested SDFG.
        for iedge in state.in_edges(node_to_replicate):
            src_node = iedge.src
            if src_node is not first_map_entry:
                # This is a connection from a node that is also replicated into the
                #  nested SDFG. We will handle it later during the output nodes, thus
                #  there is nothing to do.
                assert src_node in nodes_to_replicate
                continue

            # If we are here then we have a connection from the MapEntry node. Thus
            #  we have to read from outside data.
            outside_input_node: dace_nodes.AccessNode = state.memlet_path(iedge)[0].src
            assert isinstance(outside_input_node, dace_nodes.AccessNode)

            # If we have not seen this node yet we have to map it into the nested SDFG.
            if outside_input_node not in replicated_node_map:
                if outside_input_node.data not in nsdfg.arrays:
                    # We have not yet mapped this data into the nested SDFG. For that
                    #  we use the same descriptor as on the outside, but we have to
                    #  mark it as non-transient.
                    nested_input_desc = outside_input_node.desc(sdfg).clone()
                    nested_input_desc.transient = False

                    if isinstance(nested_input_desc, dace_data.View):
                        raise NotImplementedError("Mapping of views is not supported.")

                    nsdfg.add_datadesc(outside_input_node.data, nested_input_desc)
                    replicated_node_map[outside_input_node] = nstate.add_access(
                        outside_input_node.data
                    )
                    input_node_map[outside_input_node.data] = outside_input_node

                else:
                    # We have not found this AccessNode yet but the data it refers
                    #  to has been found. In that case we reuse the AccessNode on
                    #  the inside we have already created.
                    # NOTE: That this will slightly change the dataflow as before,
                    #   on the outside the data was accessed through two different
                    #   AccessNodes and now only one is used, also only one of them
                    #   is mapped into the nested SDFG.
                    replicated_node_map[outside_input_node] = replicated_node_map[
                        input_node_map[outside_input_node.data]
                    ]

            # We will use the "last level Memlet" to read from it.
            nstate.add_edge(
                replicated_node_map[outside_input_node],
                None,
                replicated_node_map[iedge.dst],
                iedge.dst_conn,
                dace.Memlet.from_memlet(iedge.data),
            )

    assert len(input_node_map) == len(set(input_node_map.values()))
    assert output_name is not None

    return nsdfg, input_node_map, output_name


def _get_first_map_parameter_to_second_map_mapping(
    state: dace.SDFGState,
    first_map: dace_nodes.Map,
    write_edge: dace_graph.MultiConnectorEdge[dace.Memlet],
    second_map: dace_nodes.Map,
    read_edge: dace_graph.MultiConnectorEdge[dace.Memlet],
) -> dict[str, dace_sym.SymbolicType]:
    """Expresses the values of the first Map's parameter name in terms of the second map."""
    assert first_map.get_param_num() == second_map.get_param_num()
    assert isinstance(write_edge.dst, dace_nodes.MapExit)
    assert write_edge.data.get_dst_subset(write_edge, state) is not None
    assert isinstance(read_edge.src, dace_nodes.MapEntry)
    assert read_edge.data.get_src_subset(read_edge, state) is not None

    parameter_mapping: dict[str, str] = {}

    # Currently we only consider the lower bound and nothing else, which is probably
    #  okay for the GT4Py setting.
    #  NOTE: This assumes that all the magic happens in the lower bound.
    first_map_accesses = write_edge.data.get_dst_subset(write_edge, state).min_element()
    second_map_accesses = read_edge.data.get_src_subset(read_edge, state).min_element()

    # We assume that the different dimensions are independent, i.e. each parameter is
    #  used to access exactly one dimension, this rules out accesses such as
    #  `a[i + j, i - j]`. This is okay for GT4Py. However, we have now to find out
    #  which parameter acts in which dimensions.
    first_map_param_mapping = _find_dimensions_access(first_map.params, first_map_accesses)

    # If we read the data then we use the parameters of the second Map. However, we
    #  must now figuring out to which value of the first map this corresponds to.
    for dim in range(first_map.get_param_num()):
        first_map_param = first_map_param_mapping[dim]
        first_map_access = first_map_accesses[dim]
        second_map_access = second_map_accesses[dim]

        first_map_param_uniq = dace_sym.pystr_to_symbolic(
            f"__unique_param_{dim}_{first_map_param!s}_{id(second_map_access)}"
        )
        first_map_access_uniq = first_map_access.subs({first_map_param: first_map_param_uniq})

        res = sympy.solve(sympy.Eq(first_map_access_uniq, second_map_access), first_map_param_uniq)
        if len(res) != 1:
            raise ValueError(
                f"Could not solve access in dimension {dim}: '{first_map_access} == {second_map_access}'"
            )
        parameter_mapping[first_map_param] = res[0]

    return parameter_mapping


def _find_dimensions_access(
    params: Sequence[str], accesses: Iterable[dace_sym.SymbolicType]
) -> dict[int, str]:
    """Determine in which dimension a parameter is used inside `accesses`."""
    result_mapping: dict[int, str] = {}
    aparams: set[str] = set(params)

    for i, access in enumerate(accesses):
        free_symbs = {str(fs) for fs in access.free_symbols}

        if len(free_symbs) == 0:
            # This is the case for an access `a[_i, 1]`, so there is no dimension
            #  to associate it with.
            continue

        for aparam in aparams:
            if aparam in free_symbs:
                break
        else:
            raise ValueError(
                f"Could not associate access '{access}' to any of the parameters '{aparams}'"
            )
        result_mapping[i] = aparam
        aparams.discard(aparam)

    if len(aparams) != 0:
        raise ValueError(f"Could not associate parameters: {aparams}")

    return result_mapping


def _compute_offset(
    edge_to_replace: dace_graph.MultiConnectorEdge[dace.Memlet],
    second_map_params: Sequence[str],
) -> list[dace_sym.SymbolicType]:
    """The offset needed to adjust the reading subsets in the second Map."""
    offset: list[dace_sym.SymbolicType] = []
    avail_second_map_params = set(second_map_params)
    for start, _, _ in edge_to_replace.data.src_subset:
        if str(start).isdigit():
            offset.append(dace_sym.pystr_to_symbolic("0"))
        else:
            start_symbols = {str(sym) for sym in start.free_symbols}
            used_param_symbols = start_symbols.intersection(avail_second_map_params)
            assert len(used_param_symbols) == 1

            # We do not have to use `used_param_symbols[0]` as one might suspect.
            #  Because of the additional offset has been taken care of by
            #  `_get_first_map_parameter_to_second_map_mapping()`.
            offset.append(start)
            avail_second_map_params.discard(next(iter(used_param_symbols)))

    return offset



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses

from gt4py._core import definitions as core_defs
from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.eve.extended_typing import MaybeNestedInTuple
from gt4py.next import common
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    ir_makers as im,
    misc as ir_misc,
)


DomainOrTupleThereof = tuple["DomainOrTupleThereof", ...] | common.Domain


class _DomainDeduction(NodeTranslator):
    def visit_Node(self, node: itir.Node, **kwargs):
        return None  # means we could not deduce the domain

    def visit_SymRef(
        self, node: itir.SymRef, *, sizes: dict[str, MaybeNestedInTuple[common.Domain]], **kwargs
    ) -> DomainOrTupleThereof | None:
        return sizes.get(node.id, None)

    def visit_Literal(self, node: itir.Literal, **kwargs) -> core_defs.Scalar:
        return ir_misc.value_from_literal(node)

    def visit_FunCall(self, node, **kwargs):
        args = self.generic_visit(node.args, **kwargs)

        if cpm.is_call_to(node, "tuple_get"):
            idx, expr = args
            return expr[idx]
        elif cpm.is_call_to(node, "make_tuple"):
            return tuple(args)

        return node


@dataclasses.dataclass(frozen=True)
class ReplaceGetDomainRangeWithConstants(PreserveLocationVisitor, NodeTranslator):
    """
    Replace `get_domain` calls into a tuple containing start and stop.

    Example:
        >>> from gt4py import next as gtx
        >>> KDim = common.Dimension(value="KDim", kind=common.DimensionKind.VERTICAL)
        >>> Vertex = common.Dimension(value="Vertex", kind=common.DimensionKind.HORIZONTAL)

        >>> sizes = {
        ...     "out": gtx.domain({Vertex: (0, 10), KDim: (0, 20)}),
        ... }

        >>> output_domain = im.call("unstructured_domain")(
        ...     im.named_range(
        ...         Vertex,
        ...         im.tuple_get(0, im.call("get_domain_range")("out", Vertex)),
        ...         im.tuple_get(1, im.call("get_domain_range")("out", Vertex)),
        ...     ),
        ...     im.named_range(
        ...         KDim,
        ...         im.tuple_get(0, im.call("get_domain_range")("out", KDim)),
        ...         im.tuple_get(1, im.call("get_domain_range")("out", KDim)),
        ...     ),
        ... )
        >>> ir = itir.Program(
        ...     id="test",
        ...     function_definitions=[],
        ...     params=[im.sym("inp"), im.sym("out")],
        ...     declarations=[],
        ...     body=[
        ...         itir.SetAt(
        ...             expr=im.as_fieldop("deref")(im.ref("inp")),
        ...             domain=output_domain,
        ...             target=im.ref("out"),
        ...         ),
        ...     ],
        ... )
        >>> result = ReplaceGetDomainRangeWithConstants.apply(ir, sizes=sizes)
        >>> print(result)
        test(inp, out) {
          out @ u⟨ Vertexₕ: [{0, 10}[0], {0, 10}[1][, KDimᵥ: [{0, 20}[0], {0, 20}[1][ ⟩ ← (⇑deref)(inp);
        }
    """

    @classmethod
    def apply(
        cls, program: itir.Program, sizes: dict[str, MaybeNestedInTuple[common.Domain | None]]
    ):
        return cls().visit(program, sizes=sizes)

    def visit_FunCall(self, node: itir.FunCall, **kwargs) -> itir.FunCall:
        if not cpm.is_call_to(node, "get_domain_range"):
            return self.generic_visit(node, **kwargs)

        field, dim = node.args
        assert isinstance(dim, itir.AxisLiteral)
        domain = _DomainDeduction().visit(field, sizes=kwargs["sizes"])

        if not isinstance(domain, common.Domain):
            raise ValueError(
                "Could not deduce domain of field expression. Must be a 'SymRef' or "
                "tuple expression thereof, but got:\n"
                f"'{field}'."
            )

        index = next((i for i, d in enumerate(domain.dims) if d.value == dim.value), None)
        assert index is not None, f"Dimension {dim.value} not found in {domain.dims}"

        return im.make_tuple(domain.ranges[index].start, domain.ranges[index].stop)
